# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import time
from typing import Callable

import parameterspace as ps

from blackboxopt import Evaluation, EvaluationSpecification, Objective
from blackboxopt.optimizers.random_search import RandomSearch

SPACE = ps.ParameterSpace()
SPACE.add(ps.ContinuousParameter("p1", [-1, 1]))


def _evaluation_function(eval_spec: EvaluationSpecification) -> Evaluation:
    loss = eval_spec.configuration["p1"] ** 2
    return eval_spec.create_evaluation(objectives={"loss": loss})


def limit_with_max_evaluations(run_optimization_loop: Callable, loop_kwargs: dict):
    evaluations = run_optimization_loop(
        RandomSearch(SPACE, [Objective("loss", False)], max_steps=10),
        _evaluation_function,
        max_evaluations=8,
        **loop_kwargs,
    )

    assert len(evaluations) == 8
    assert all([not e.all_objectives_none for e in evaluations])


def limit_with_loop_timeout(run_optimization_loop: Callable, loop_kwargs: dict):
    class SlowRandomSearch(RandomSearch):
        def generate_evaluation_specification(self) -> EvaluationSpecification:
            time.sleep(1)
            return super().generate_evaluation_specification()

    evaluations = run_optimization_loop(
        SlowRandomSearch(SPACE, [Objective("loss", False)], max_steps=10),
        _evaluation_function,
        timeout_s=3.0,
        **loop_kwargs,
    )
    assert len(evaluations) < 4
    assert all([not e.all_objectives_none for e in evaluations])


def reporting_user_info(run_optimization_loop: Callable, loop_kwargs: dict):
    def __evaluation_function(eval_spec):
        return eval_spec.create_evaluation(
            objectives={"loss": 1.0}, user_info={"user": "info"}
        )

    max_steps = 10
    evaluations = run_optimization_loop(
        RandomSearch(SPACE, [Objective("loss", False)], max_steps=max_steps),
        __evaluation_function,
        **loop_kwargs,
    )

    assert len(evaluations) == max_steps
    assert all([not e.all_objectives_none for e in evaluations])
    assert all([e.user_info is not None for e in evaluations])


ALL_REFERENCE_TESTS = [
    limit_with_max_evaluations,
    limit_with_loop_timeout,
    reporting_user_info,
]



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import json
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from datetime import datetime
from typing import Any, Dict, Mapping, Optional

import numpy as np


def _datetime_now_timestamp():
    """Wrapper to allow use as default factory for dataclass fields."""
    return datetime.now().timestamp()


@dataclass
class EvaluationSpecification(Mapping[str, Any]):
    configuration: dict = field(
        metadata={"Description": "The configuration to be evaluated next."}
    )

    settings: dict = field(
        default_factory=dict,
        metadata={
            "Description": "Additional settings like the fidelity or target task."
        },
    )

    optimizer_info: dict = field(
        default_factory=dict,
        metadata={"Description": "Information about and for internal optimizer state."},
    )

    created_unixtime: float = field(
        default_factory=_datetime_now_timestamp,
        metadata={"Description": "Creation time of the evaluation specificiation."},
    )

    context: Optional[Dict[str, Any]] = field(
        default=None,
        metadata={
            "Description": "Contextual information is what you can determine but not "
            + "influence, like the environmental temperature."
        },
    )

    def keys(self):
        return self.__dataclass_fields__.keys()  # pylint: disable=no-member

    def create_evaluation(
        self,
        objectives: Dict[str, Optional[float]],
        constraints: Optional[Dict[str, Optional[float]]] = None,
        user_info: Optional[dict] = None,
        stacktrace: Optional[str] = None,
        finished_unixtime: Optional[float] = None,
    ):
        """Create a blackboxopt.Evaluation based on this evaluation specification.

        Args:
            objectives: For each objective name the respective value.
            constraints: For each constraint name the float value indicates how much the
                constraint was satisfied, with negative values implying a violated and
                positive values indicating a satisfied constraint.
            user_info: Miscellaneous information provided by the user.
            stacktrace: The stacktrace in case an unhandled exception occurred inside
                the evaluation function.
            finished_unixtime: Timestamp at completion of this evaluation. If none is
                provided, the current time is used.
        """
        evaluation = Evaluation(
            objectives=objectives,
            constraints=constraints,
            user_info=user_info,
            stacktrace=stacktrace,
            **self,
        )

        # Data class default factories like in this case time.time are only triggered
        # when the argument is not provided, so in case of it being None we can't just
        # pass the argument value in, because it would set it to None instead of
        # triggering the default factory for the current time.
        if finished_unixtime is not None:
            evaluation.finished_unixtime = finished_unixtime

        return evaluation

    def __getitem__(self, key):
        if key not in self.__dataclass_fields__:  # pylint: disable=no-member
            raise KeyError(
                f"Only dataclass fields are accessible via __getitem__, '{key}' is not."
            )

        return deepcopy(getattr(self, key))

    def __iter__(self):
        return self.__dataclass_fields__.__iter__  # pylint: disable=no-member

    def __len__(self):
        return self.__dataclass_fields__.__len__  # pylint: disable=no-member

    def to_json(self, **json_dump_kwargs):
        return json.dumps(asdict(self), **json_dump_kwargs)

    def to_dict(self):
        return self.__dict__


@dataclass
class _EvaluationBase:
    """Helper dataclass to allow the Evaluation class to have attributes with defaults
    while still having attributes without default values. To make this happen, the
    non-default attributes need to be defined / inherited before the ones with defaults.

    Attributes:
        objectives: [description]

    Raises:
        ValueError: [description]
    """

    objectives: Dict[str, Optional[float]]

    def __post_init__(self):
        available_objective_values = np.array(
            [o for o in self.objectives.values() if o is not None], dtype=float
        )
        if not np.isfinite(available_objective_values).all():
            raise ValueError(
                f"Objective values contain non-finite values: {self.objectives}\n"
                + "Please use None to indicate missing objective evaluations."
            )


@dataclass
class Evaluation(EvaluationSpecification, _EvaluationBase):
    """An evaluated specification with a timestamp indicating the time of the
    evaluation, and a result dictionary for all objective values.

    NOTE: `NaN` is not allowed as an objective value, use `None` instead.
    """

    constraints: Optional[Dict[str, Optional[float]]] = field(
        default=None,
        metadata={
            "Description": "For each constraint name the float value indicates "
            + "how much the constraint was satisfied, with negative values implying "
            + "a violated and positive values indicating a satisfied constraint."
        },
    )

    finished_unixtime: float = field(
        default_factory=_datetime_now_timestamp,
        metadata={"Description": "Timestamp at completion of this evaluation."},
    )

    stacktrace: Optional[str] = field(
        default=None,
        metadata={
            "Description": "The stacktrace in case an unhandled exception occurred "
            + "inside the evaluation function."
        },
    )

    user_info: Optional[dict] = field(
        default=None,
        metadata={"Description": "Miscellaneous information provided by the user."},
    )

    def get_specification(
        self, reset_created_unixtime: bool = False
    ) -> EvaluationSpecification:
        """Get the evaluation specifiation for which this result was evaluated."""
        eval_spec_kwargs = deepcopy(
            dict(
                configuration=self.configuration,
                settings=self.settings,
                optimizer_info=self.optimizer_info,
                context=self.context,
            )
        )

        if reset_created_unixtime:
            return EvaluationSpecification(
                created_unixtime=_datetime_now_timestamp(), **eval_spec_kwargs
            )

        return EvaluationSpecification(
            created_unixtime=self.created_unixtime, **eval_spec_kwargs
        )

    @property
    def any_objective_none(self) -> bool:
        return any([v is None for v in self.objectives.values()])

    @property
    def all_objectives_none(self) -> bool:
        return all([v is None for v in self.objectives.values()])



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

try:
    from scipy.stats.qmc import Sobol
except ImportError as e:
    raise ImportError(
        "Unable to import SpaceFilling optimizer specific dependencies. "
        + "Make sure to install blackboxopt[space-fill]"
    ) from e
from blackboxopt.base import MultiObjectiveOptimizer, Objective, ParameterSpace
from blackboxopt.evaluation import EvaluationSpecification


class SpaceFilling(MultiObjectiveOptimizer):
    """Sobol sequence based, space filling optimizer.

    Args:
        search_space: The search space to optimize
        objectives: The objectives of the optimization
        seed: The sobol sequence is Owen scrambled and can be seeded for reproducibility
    """

    def __init__(
        self,
        search_space: ParameterSpace,
        objectives: List[Objective],
        seed: int = None,
    ) -> None:
        super().__init__(search_space=search_space, objectives=objectives, seed=seed)
        self.sobol = Sobol(d=len(self.search_space), scramble=True, seed=seed)

    def generate_evaluation_specification(self) -> EvaluationSpecification:
        vector = self.sobol.random().flatten()
        configuration = self.search_space.from_numerical(vector)
        return EvaluationSpecification(configuration=configuration)



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0
#
# This source code is derived from HpBandSter 0.7.4
#  https://github.com/automl/HpBandSter
# Copyright (c) 2017-2018, ML4AAD, licensed under the BSD-3 license,
# cf. 3rd-party-licenses.txt file in the root directory of this source tree.
# Changes include:
#     - integration into the blackboxopt API
#     - docstrings and type hints

import logging

from parameterspace import ParameterSpace

from blackboxopt import Objective

try:
    from blackboxopt.optimizers.staged.configuration_sampler import RandomSearchSampler
    from blackboxopt.optimizers.staged.hyperband import create_hyperband_iteration
    from blackboxopt.optimizers.staged.iteration import StagedIteration
    from blackboxopt.optimizers.staged.optimizer import StagedIterationOptimizer
except ImportError as e:
    raise ImportError(
        "Unable to import Hyperband optimizer specific dependencies. "
        + "Make sure to install blackboxopt[hyperband]"
    ) from e


class Hyperband(StagedIterationOptimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objective: Objective,
        min_fidelity: float,
        max_fidelity: float,
        num_iterations: int,
        eta: float = 3.0,
        seed: int = None,
        logger: logging.Logger = None,
    ):
        """Implementation of Hyperband as proposed in

        Li, L., Jamieson, K., DeSalvo, G., Rostamizadeh, A., & Talwalkar, A. (2016).
        Hyperband: A novel bandit-based approach to hyperparameter optimization.
        arXiv preprint arXiv:1603.06560.

        Args:
            search_space: [description]
            objective: [description]
            min_fidelity: The smallest fidelity value that is still meaningful.
                Must be strictly greater than zero!
            max_fidelity: The largest fidelity value used during the optimization.
                Must not be smaller than `min_fidelity`
            num_iterations: [description]
            eta: Scaling parameter to control the aggressiveness of Hyperband's racing.
            seed: [description]
            logger: [description]
        """
        self.config_sampler = RandomSearchSampler(search_space)
        self.min_fidelity = min_fidelity
        self.max_fidelity = max_fidelity
        self.eta = eta

        super().__init__(
            search_space=search_space,
            objective=objective,
            num_iterations=num_iterations,
            seed=seed,
            logger=logger,
        )

    def _create_new_iteration(self, iteration_index: int) -> StagedIteration:
        """Optimizer specific way to create a new
        `blackboxopt.optimizer.staged.iteration.StagedIteration` object
        """
        return create_hyperband_iteration(
            iteration_index,
            self.min_fidelity,
            self.max_fidelity,
            self.eta,
            self.config_sampler,
            self.objective,
            self.logger,
        )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import hashlib
import pickle
from itertools import compress
from typing import Dict, Iterable, List, Optional, Sequence

import numpy as np

from blackboxopt.base import Objective
from blackboxopt.evaluation import Evaluation


def get_loss_vector(
    known_objectives: Sequence[Objective],
    reported_objectives: Dict[str, Optional[float]],
    none_replacement: float = float("NaN"),
) -> List[float]:
    """Convert reported objectives into a vector of known objectives.

    Args:
        known_objectives: A sequence of objectives with names and directions
            (whether greate is better). The order of the objectives dictates the order
            of the returned loss values.
        reported_objectives: A dictionary with the objective value for each of the known
            objectives' names.
        none_replacement: The value to use for missing objective values that are `None`

    Returns:
        A list of loss values.
    """
    losses = []
    for objective in known_objectives:
        objective_value = reported_objectives[objective.name]
        if objective_value is None:
            losses.append(none_replacement)
        elif objective.greater_is_better:
            losses.append(-1.0 * objective_value)
        else:
            losses.append(objective_value)

    return losses


def mask_pareto_efficient(losses: np.ndarray):
    """For a given array of objective values where lower values are considered better
    and the dimensions are samples x objectives, return a mask that is `True` for all
    pareto efficient values.

    NOTE: The result marks multiple occurrences of the same point all as pareto
    efficient.
    """
    is_efficient = np.ones(losses.shape[0], dtype=bool)
    for i, c in enumerate(losses):
        if not is_efficient[i]:
            continue

        # Keep any point with a lower cost or when they are the same
        efficient = np.any(losses[is_efficient] < c, axis=1)
        duplicates = np.all(losses[is_efficient] == c, axis=1)
        is_efficient[is_efficient] = np.logical_or(efficient, duplicates)

    return is_efficient


def filter_pareto_efficient(
    evaluations: List[Evaluation], objectives: List[Objective]
) -> List[Evaluation]:
    """Filter pareto efficient evaluations with respect to given objectives."""
    losses = np.array(
        [
            get_loss_vector(
                known_objectives=objectives, reported_objectives=e.objectives
            )
            for e in evaluations
        ]
    )

    pareto_efficient_mask = mask_pareto_efficient(losses)

    return list(compress(evaluations, pareto_efficient_mask))


def sort_evaluations(evaluations: Iterable[Evaluation]) -> Iterable[Evaluation]:
    return sorted(
        evaluations,
        key=lambda e: hashlib.md5(
            pickle.dumps(
                [
                    sorted(e.configuration.items()),
                    sorted(e.objectives.items()),
                    sorted(e.settings.items()),
                    sorted(e.constraints.items()) if e.constraints else {},
                    sorted(e.context.items()) if e.context else {},
                ]
            )
        ).hexdigest(),
    )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import logging
import time
from typing import Callable, List, Set, Union

try:
    import dask.distributed as dd
except ImportError as e:
    raise ImportError(
        "Unable to import Dask Distributed specific dependencies. "
        + "Make sure to install blackboxopt[dask]"
    ) from e

from blackboxopt import (
    Evaluation,
    EvaluationSpecification,
    OptimizationComplete,
    OptimizerNotReady,
)
from blackboxopt.base import (
    MultiObjectiveOptimizer,
    Objective,
    SingleObjectiveOptimizer,
)
from blackboxopt.optimization_loops.utils import (
    evaluation_function_wrapper,
    init_max_evaluations_with_limit_logging,
)


class MinimalDaskScheduler:
    def __init__(
        self,
        dask_client: dd.Client,
        objectives: List[Objective],
        logger: logging.Logger,
    ):
        self.client = dask_client
        self.objectives = objectives
        self.logger = logger
        self._not_done_futures: Set = set()

    def shutdown(self):
        return self.client.shutdown()

    def has_capacity(self):
        idle = [len(task_id) == 0 for task_id in self.client.processing().values()]
        return sum(idle)

    def has_running_jobs(self):
        return len(self._not_done_futures) > 0

    def submit(
        self,
        eval_function: Callable[[EvaluationSpecification], Evaluation],
        eval_spec: EvaluationSpecification,
    ):
        f = self.client.submit(
            evaluation_function_wrapper,
            evaluation_function=eval_function,
            evaluation_specification=eval_spec,
            objectives=self.objectives,
            catch_exceptions_from_evaluation_function=True,
            logger=self.logger,
        )
        f.bbo_eval_spec = eval_spec
        self._not_done_futures.add(f)

    def check_for_results(self, timeout_s: float = 5.0) -> List[Evaluation]:
        try:
            all_futures = dd.wait(
                self._not_done_futures, timeout=timeout_s, return_when="FIRST_COMPLETED"
            )

            evaluations: List[Evaluation] = []
            for f in all_futures.done:
                if f.status == "error":
                    evaluation = Evaluation(
                        objectives={o.name: None for o in self.objectives},
                        stacktrace=str(f.traceback()),
                        **f.bbo_eval_spec
                    )
                else:
                    evaluation = f.result()

                self.logger.debug(evaluation.__dict__)
                evaluations.append(evaluation)

            self._not_done_futures = all_futures.not_done
            return evaluations

        except dd.TimeoutError:
            return []


def run_optimization_loop(
    optimizer: Union[SingleObjectiveOptimizer, MultiObjectiveOptimizer],
    evaluation_function: Callable[[EvaluationSpecification], Evaluation],
    dask_client: dd.Client,
    timeout_s: float = float("inf"),
    max_evaluations: int = None,
    logger: logging.Logger = None,
) -> List[Evaluation]:
    """Convenience wrapper for an optimization loop that uses Dask to parallelize
    optimization until a given timeout or maximum number of evaluations is reached.

    This already handles signals from the optimizer in case there is no evaluation
    specification available yet.

    Args:
        optimizer: The blackboxopt optimizer to run.
        dask_client: A Dask Distributed client that is configured with workers.
        evaluation_function: The function that is called with configuration, settings
            and optimizer info dictionaries as arguments like provided by an evaluation
            specification.
            This is the function that encapsulates the actual execution of
            a parametrized experiment (e.g. ML model training) and should return a
            `blackboxopt.Evaluation` as a result.
        timeout_s: If given, the optimization loop will terminate after the first
            optimization step that exceeded the timeout (in seconds). Defaults to inf.
        max_evaluations: If given, the optimization loop will terminate after the given
            number of steps. Defaults to None.
        logger: The logger to use for logging progress. Defaults to None.

    Returns:
        List of evluation specification and result for all evaluations.
    """
    logger = logging.getLogger("blackboxopt") if logger is None else logger

    objectives = (
        optimizer.objectives
        if isinstance(optimizer, MultiObjectiveOptimizer)
        else [optimizer.objective]
    )
    evaluations: List[Evaluation] = []

    dask_scheduler = MinimalDaskScheduler(
        dask_client=dask_client, objectives=objectives, logger=logger
    )

    _max_evaluations = init_max_evaluations_with_limit_logging(
        max_evaluations=max_evaluations, timeout_s=timeout_s, logger=logger
    )

    n_eval_specs = 0
    start = time.time()
    while time.time() - start < timeout_s and n_eval_specs < _max_evaluations:
        if dask_scheduler.has_capacity():
            try:
                eval_spec = optimizer.generate_evaluation_specification()
                dask_scheduler.submit(evaluation_function, eval_spec)
                n_eval_specs += 1
                continue

            except OptimizerNotReady:
                logger.info("Optimizer is not ready yet; will retry after short pause.")

            except OptimizationComplete:
                logger.info("Optimization is complete")
                break

        new_evaluations = dask_scheduler.check_for_results(timeout_s=20)
        optimizer.report(new_evaluations)
        evaluations.extend(new_evaluations)

    while dask_scheduler.has_running_jobs():
        new_evaluations = dask_scheduler.check_for_results(timeout_s=20)
        optimizer.report(new_evaluations)
        evaluations.extend(new_evaluations)

    return evaluations



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Tuple

from parameterspace import ParameterSpace

from blackboxopt import Evaluation


class StagedIterationConfigurationSampler:
    """Base class for sampling new configurations inside a StagedIterationOpitimzer."""

    @abc.abstractmethod
    def sample_configuration(self) -> Tuple[dict, dict]:
        """Pick the next configuration.

        Returns:
            The configuration to be evaluated,
            Additional information that will be added to the `optimizer_info` dict.
        """

    @abc.abstractmethod
    def digest_evaluation(self, evaluation: Evaluation):
        """Register the result of an evaluation."""


class RandomSearchSampler(StagedIterationConfigurationSampler):
    def __init__(self, search_space: ParameterSpace):
        self.search_space = search_space

    def sample_configuration(self) -> Tuple[dict, dict]:
        return self.search_space.sample(), {}

    def digest_evaluation(self, evaluation: Evaluation):
        """Random Search is stateless and does nothing with finished evaluations."""



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

from blackboxopt import (
    EvaluationSpecification,
    Objective,
    OptimizationComplete,
    ParameterSpace,
)
from blackboxopt.base import MultiObjectiveOptimizer


class RandomSearch(MultiObjectiveOptimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objectives: List[Objective],
        max_steps: int,
        seed: int = None,
    ) -> None:
        """Randomly sample up to `max_steps` configurations from the given search space.

        Args:
            search_space: Space to search
            objectives: The objectives of the optimization.
            max_steps: Max number of evaluation specifications the optimizer generates
                before raising `OptimizationComplete`
            seed: Optional number to seed the random number generator with.
                Defaults to None.
        """
        super().__init__(search_space=search_space, objectives=objectives, seed=seed)

        self.max_steps: int = max_steps
        self.n_steps: int = 0

    def generate_evaluation_specification(self) -> EvaluationSpecification:
        """[summary]

        Raises:
            OptimizationComplete: Raised if the optimizer's `max_steps` are reached.

        Returns:
            [description]
        """
        if self.n_steps >= self.max_steps:
            raise OptimizationComplete()

        eval_spec = EvaluationSpecification(
            configuration=self.search_space.sample(),
            settings={},
            optimizer_info={"step": self.n_steps},
        )
        self.n_steps += 1

        return eval_spec



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

"""Tests that can be imported and used to test optimizer implementations against this
packages blackbox optimizer interface."""

import random
from typing import Dict, List, Optional, Type, Union

import numpy as np
import parameterspace as ps

from blackboxopt import Evaluation, Objective, OptimizationComplete, Optimizer
from blackboxopt.base import (
    EvaluationsError,
    MultiObjectiveOptimizer,
    SingleObjectiveOptimizer,
)


def _initialize_optimizer(
    optimizer_class,
    optimizer_kwargs: dict,
    objective: Objective,
    objectives: List[Objective],
    space: Optional[ps.ParameterSpace] = None,
    seed=42,
) -> Optimizer:
    if space is None:
        space = ps.ParameterSpace()
        space.add(ps.IntegerParameter("p1", bounds=(1, 32), transformation="log"))
        space.add(ps.ContinuousParameter("p2", (-2, 2)))
        space.add(ps.ContinuousParameter("p3", (0, 1)))
        space.add(ps.CategoricalParameter("p4", [True, False]))
        space.add(ps.OrdinalParameter("p5", ("small", "medium", "large")))

    if issubclass(optimizer_class, MultiObjectiveOptimizer):
        return optimizer_class(space, objectives, seed=seed, **optimizer_kwargs)

    if issubclass(optimizer_class, SingleObjectiveOptimizer):
        return optimizer_class(space, objective, seed=seed, **optimizer_kwargs)

    return optimizer_class(space, seed=seed, **optimizer_kwargs)


def optimize_single_parameter_sequentially_for_n_max_evaluations(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
    n_max_evaluations: int = 20,
):
    """[summary]

    Args:
        optimizer_class: [description]
        optimizer_kwargs: [description]
        n_max_evaluations: [description]

    Returns:
        [description]
    """

    def quadratic_function(p1):
        return p1**2

    assert issubclass(optimizer_class, Optimizer), (
        "The default test suite is only applicable for implementations of "
        "blackboxopt.base.Optimizer"
    )

    optimizer = _initialize_optimizer(
        optimizer_class,
        optimizer_kwargs,
        objective=Objective("loss", False),
        objectives=[Objective("loss", False), Objective("score", True)],
    )

    eval_spec = optimizer.generate_evaluation_specification()

    if issubclass(optimizer_class, MultiObjectiveOptimizer):
        evaluation = eval_spec.create_evaluation(
            objectives={"loss": None, "score": None}
        )
    else:
        evaluation = eval_spec.create_evaluation(
            objectives={"loss": None}, constraints={"constraint": None}
        )
    optimizer.report(evaluation)

    for _ in range(n_max_evaluations):

        try:
            eval_spec = optimizer.generate_evaluation_specification()
        except OptimizationComplete:
            break

        loss = quadratic_function(p1=eval_spec.configuration["p1"])
        if issubclass(optimizer_class, MultiObjectiveOptimizer):
            evaluation_result = {"loss": loss, "score": -loss}
        else:
            evaluation_result = {"loss": loss}

        evaluation = eval_spec.create_evaluation(
            objectives=evaluation_result, constraints={"constraint": 10.0}
        )
        optimizer.report(evaluation)


def is_deterministic_with_fixed_seed_and_larger_space(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if optimizer is deterministic.

    Initialize the optimizer twice with the same parameter space and a fixed seed.
    For each optimizer run optimization loop `n_evaluations` times, namely,
    get an evaluation specification and report a placeholder result back.
    The list of configurations should be equal for both optimizers.

    This tests covers multiple parameter types by using a mixed search space.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """
    n_evaluations = 5
    losses = [0.1, 0.2, 0.3, 0.4, 0.5]

    run_0_configs: List[Evaluation] = []
    run_1_configs: List[Evaluation] = []

    for run_configs in [run_0_configs, run_1_configs]:
        opt = _initialize_optimizer(
            optimizer_class,
            optimizer_kwargs,
            objective=Objective("loss", False),
            objectives=[Objective("loss", False)],
            seed=42,
        )

        for i in range(n_evaluations):
            es = opt.generate_evaluation_specification()
            evaluation = es.create_evaluation(
                objectives={"loss": losses[i]}, constraints={"constraint": 10.0}
            )
            opt.report(evaluation)

            run_configs.append(evaluation.configuration)

    assert len(run_0_configs) == n_evaluations
    assert run_0_configs == run_1_configs


def is_deterministic_when_reporting_shuffled_evaluations(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if determinism isn't affected by the order of initially reported data.

    Repeatedly initialize the optimizer with the same parameter space and a fixed seed.
    Report a set of initial evaluations in randomized order as initial data. Start
    optimizing and check if the generated configurations for all optimizers are equal.

    By doing multiple evaluations, this tests covers effects that become visible after
    a while, e.g. only after stages got completed in staged iteration samplers.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """

    space = ps.ParameterSpace()
    space.add(ps.ContinuousParameter("p1", (0, 1)))

    def _run_experiment_1d(es):
        x = es.configuration["p1"]
        _x = np.copy(np.atleast_2d(x))
        params = np.array([0.75, 0.0, -10.0, 0.0, 0.0])
        y = np.polyval(params, _x)
        return float(np.squeeze(y))

    runs: Dict[int, Dict] = {0: {}, 1: {}}
    for run_idx, run in runs.items():
        run["evaluations"] = []
        opt = _initialize_optimizer(
            optimizer_class,
            optimizer_kwargs,
            objective=Objective("loss", False),
            objectives=[Objective("loss", False)],
            space=space,
            seed=0,
        )

        # Report initial data in different order
        eval_specs = [opt.generate_evaluation_specification() for _ in range(5)]
        run["initial_evaluations"] = [
            es.create_evaluation(
                objectives={"loss": _run_experiment_1d(es)},
                constraints={"constraint": 10.0},
            )
            for es in eval_specs
        ]
        random.seed(run_idx)
        random.shuffle(run["initial_evaluations"])
        opt.report(run["initial_evaluations"])

        # Start optimizing
        for _ in range(5):
            es = opt.generate_evaluation_specification()
            evaluation = es.create_evaluation(
                objectives={"loss": _run_experiment_1d(es)},
                constraints={"constraint": 10.0},
            )
            opt.report(evaluation)
            run["evaluations"].append(evaluation)

    initial_configs_run_0 = [e.configuration for e in runs[0]["initial_evaluations"]]
    initial_configs_run_1 = [e.configuration for e in runs[1]["initial_evaluations"]]

    configs_run_0 = [e.configuration for e in runs[0]["evaluations"]]
    configs_run_1 = [e.configuration for e in runs[1]["evaluations"]]

    assert initial_configs_run_0 != initial_configs_run_1
    assert configs_run_0 == configs_run_1


def handles_reporting_evaluations_list(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if optimizer's report method can process an iterable of evaluations.

    All optimizers should be able to allow reporting batches of evaluations. It's up to
    the optimizer's implementation, if evaluations in a batch are processed
    one by one like if they were reported individually, or if a batch is handled
    differently.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """
    opt = _initialize_optimizer(
        optimizer_class,
        optimizer_kwargs,
        objective=Objective("loss", False),
        objectives=[Objective("loss", False)],
        seed=42,
    )
    evaluations = []
    for _ in range(3):
        es = opt.generate_evaluation_specification()
        evaluation = es.create_evaluation(
            objectives={"loss": 0.42}, constraints={"constraint": 10.0}
        )
        evaluations.append(evaluation)

    opt.report(evaluations)


def raises_evaluation_error_when_reporting_unknown_objective(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if optimizer's report method raises exception in case objective is unknown.

    Also make sure that the faulty evaluations (and only those) are included in the
    exception.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """
    opt = _initialize_optimizer(
        optimizer_class,
        optimizer_kwargs,
        objective=Objective("loss", False),
        objectives=[Objective("loss", False)],
    )
    es_1 = opt.generate_evaluation_specification()
    es_2 = opt.generate_evaluation_specification()
    es_3 = opt.generate_evaluation_specification()

    # NOTE: The following is not using pytest.raises because this would add pytest as
    #       a regular dependency to blackboxopt.
    try:
        evaluation_1 = es_1.create_evaluation(
            objectives={"loss": 1}, constraints={"constraint": 10.0}
        )
        evaluation_2 = es_2.create_evaluation(
            objectives={"unknown_objective": 2}, constraints={"constraint": 10.0}
        )
        evaluation_3 = es_3.create_evaluation(
            objectives={"loss": 4}, constraints={"constraint": 10.0}
        )
        opt.report([evaluation_1, evaluation_2, evaluation_3])

        raise AssertionError(
            f"Optimizer {optimizer_class} did not raise an ObjectivesError when a "
            + "result including an unknown objective name was reported."
        )

    except EvaluationsError as exception:
        invalid_evaluations = [e for e, _ in exception.evaluations_with_errors]
        assert len(invalid_evaluations) == 1
        assert evaluation_2 in invalid_evaluations


def respects_fixed_parameter(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if optimizer's generated evaluation specifications contain the values
    a parameter in the search space was fixed to.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """
    space = ps.ParameterSpace()
    space.add(ps.ContinuousParameter("my_fixed_param", (-10.0, 200.0)))
    space.add(ps.ContinuousParameter("x", (-2.0, 2.0)))

    fixed_value = 1.0
    space.fix(my_fixed_param=fixed_value)
    opt = _initialize_optimizer(
        optimizer_class,
        optimizer_kwargs,
        objective=Objective("loss", False),
        objectives=[Objective("loss", False)],
        space=space,
    )
    for _ in range(5):
        es = opt.generate_evaluation_specification()
        assert es.configuration["my_fixed_param"] == fixed_value
        opt.report(
            es.create_evaluation(
                objectives={"loss": es.configuration["x"] ** 2},
                constraints={"constraint": 10.0},
            )
        )


def handles_conditional_space(
    optimizer_class: Union[
        Type[SingleObjectiveOptimizer], Type[MultiObjectiveOptimizer]
    ],
    optimizer_kwargs: dict,
):
    """Check if optimizer handles conditional i.e. hierarchical search spaces.

    Args:
        optimizer_class: Optimizer to test.
        optimizer_kwargs: Expected to contain additional arguments for initializing
            the optimizer. (`search_space` and `objective(s)` are set automatically
            by the test.)

    Returns:
        `True` if the test is passed.
    """
    space = ps.ParameterSpace()
    space.add(ps.CategoricalParameter("optimizer", ("adam", "sgd")))
    space.add(ps.ContinuousParameter("lr", (0.0001, 0.1), transformation="log"))
    space.add(
        ps.ContinuousParameter("momentum", (0.0, 1.0)),
        lambda optimizer: optimizer == "sgd",
    )

    opt = _initialize_optimizer(
        optimizer_class,
        optimizer_kwargs,
        objective=Objective("loss", False),
        objectives=[Objective("loss", False)],
        space=space,
    )

    for _ in range(10):
        es = opt.generate_evaluation_specification()
        dummy_loss = es.configuration.get("momentum", 1.0) * es.configuration["lr"] ** 2
        opt.report(
            es.create_evaluation(
                objectives={"loss": dummy_loss}, constraints={"constraint": 10.0}
            )
        )


ALL_REFERENCE_TESTS = [
    optimize_single_parameter_sequentially_for_n_max_evaluations,
    is_deterministic_with_fixed_seed_and_larger_space,
    is_deterministic_when_reporting_shuffled_evaluations,
    handles_reporting_evaluations_list,
    raises_evaluation_error_when_reporting_unknown_objective,
    respects_fixed_parameter,
    handles_conditional_space,
]



__version__ = "4.7.1"

from parameterspace import ParameterSpace

from .base import (
    ConstraintsError,
    ContextError,
    EvaluationsError,
    Objective,
    ObjectivesError,
    OptimizationComplete,
    Optimizer,
    OptimizerNotReady,
)
from .evaluation import Evaluation, EvaluationSpecification
from .utils import sort_evaluations



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import datetime
import logging
import traceback
from typing import Callable, List

from blackboxopt import Evaluation, EvaluationSpecification, Objective
from blackboxopt.base import ObjectivesError, raise_on_unknown_or_incomplete


class EvaluationFunctionError(ValueError):
    """Raised on errors originating from the user defined evaluation function."""

    def __init__(self, evaluation_specification: EvaluationSpecification):
        self.message = (
            "An error occurred when attempting to call the user specified evaluation "
            "function with the specification below. Please check the cause of this "
            "exception in the output further up for the original stacktrace.\n"
            f"{evaluation_specification}"
        )
        self.evaluation_specification = evaluation_specification


def init_max_evaluations_with_limit_logging(
    timeout_s: float, logger: logging.Logger, max_evaluations: int = None
) -> float:
    """[summary]

    Args:
        timeout_s: [description]
        logger: [description]
        max_evaluations: [description]

    Returns:
        [description]
    """
    if max_evaluations:
        logger.info(
            "Starting optimization run. Stops when complete or "
            + f"{max_evaluations} evaluations reached."
        )
        return float(max_evaluations)

    if timeout_s == float("inf"):
        logger.info("Starting optimization run. Stops when complete.")
    else:
        timeout_pretty = datetime.timedelta(seconds=timeout_s)
        logger.info(
            "Starting optimization run. Stops when complete or "
            + f"{timeout_pretty} passed."
        )

    return float("inf")


def evaluation_function_wrapper(
    evaluation_function: Callable[[EvaluationSpecification], Evaluation],
    evaluation_specification: EvaluationSpecification,
    objectives: List[Objective],
    catch_exceptions_from_evaluation_function: bool,
    logger: logging.Logger,
) -> Evaluation:
    """Wrapper for evaluation functions. The evaluation result returned by the
    evaluation function is checked to contain all relevant objectives. An empty
    evaluation with a stacktrace is reported to the optimizer in case an unhandled
    Exception occurrs during the evaluation function call when
    `catch_exceptions_from_evaluation_function` is set to `True`, otherwise an
    `EvaluationFunctionError` is raised based on the original exception.
    """
    try:
        evaluation = evaluation_function(evaluation_specification)
    except Exception as e:
        if not catch_exceptions_from_evaluation_function:
            raise EvaluationFunctionError(evaluation_specification) from e

        stacktrace = traceback.format_exc()

        logger.warning(
            "Reporting FAILURE due to unhandled error in evaluation function. See "
            + "DEBUG log level output or evaluation.stacktrace for details. "
            + "Alternatively, disable automated exception handling by setting "
            + "catch_exceptions_from_evaluation_function=False to exit on errors."
        )
        logger.debug(stacktrace)

        evaluation = evaluation_specification.create_evaluation(
            stacktrace=stacktrace, objectives={o.name: None for o in objectives}
        )

    raise_on_unknown_or_incomplete(
        exception=ObjectivesError,
        known=[o.name for o in objectives],
        reported=evaluation.objectives.keys(),
    )

    return evaluation



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import copy
import logging
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional, Tuple
from uuid import UUID, uuid4

import numpy as np

from blackboxopt import Evaluation, EvaluationSpecification, Objective
from blackboxopt.optimizers.staged.configuration_sampler import (
    StagedIterationConfigurationSampler,
)


@dataclass
class Datum:
    """Small container for bookkeeping only."""

    config_key: Tuple[int, int, int]
    status: str
    loss: float = float("NaN")


class StagedIteration:
    def __init__(
        self,
        iteration: int,
        num_configs: List[int],
        fidelities: List[float],
        config_sampler: StagedIterationConfigurationSampler,
        config_promotion_function: Callable,
        objective: Objective,
        logger: logging.Logger = None,
    ):
        """Base class for iterations that compare configurations at different
        fidelities and race them as in SuccessiveHalving or Hyperband.

        Args:
            iteration: Index of this iteration.
            num_configs: Number of configurations in each stage.
            fidelities: The fidelity for each stage. Must have the same length as
                `num_configs'.
            config_sampler: Configuration Sampler object that suggests a new
                configuration for evaluation given a fidelity.
            config_promotion_function: Function that decides which configurations are
                promoted. Check
                `blackboxopt.optimizers.utils.staged_iteration.greedy_promotion` for
                the signature.
            objective: The objective of the optimization.
            logger: A standard logger to which some debug output might be written.
        """
        assert len(fidelities) == len(
            num_configs
        ), "Please specify the number of configuration and the fidelities."
        self.logger = logging.getLogger("blackboxopt") if logger is None else logger
        self.iteration = iteration
        self.fidelities = fidelities
        self.num_configs = num_configs
        self.config_sampler = config_sampler
        self.config_promotion_function = config_promotion_function
        self.objective = objective
        self.current_stage = 0
        self.evaluation_data: List[List[Datum]] = [[]]
        self.eval_specs: Dict[Tuple[int, int, int], EvaluationSpecification] = {}
        self.pending_evaluations: Dict[UUID, int] = {}
        self.finished = False

    def generate_evaluation_specification(self) -> Optional[EvaluationSpecification]:
        """Pick the next evaluation specification with a budget i.e. fidelity to run.

        Returns:
            [description]
        """
        if self.finished:
            return None

        # try to find a queued entry first
        for i, d in enumerate(self.evaluation_data[self.current_stage]):
            if d.status == "QUEUED":
                es = copy.deepcopy(self.eval_specs[d.config_key])
                es.settings["fidelity"] = self.fidelities[self.current_stage]
                d.status = "RUNNING"
                self.pending_evaluations[es.optimizer_info["id"]] = i
                return es

        # sample a new configuration if there are empty slots to be filled
        if (
            len(self.evaluation_data[self.current_stage])
            < self.num_configs[self.current_stage]
        ):
            conf_key = (
                self.iteration,
                self.current_stage,
                len(self.evaluation_data[self.current_stage]),
            )
            conf, opt_info = self.config_sampler.sample_configuration()
            opt_info.update({"configuration_key": conf_key, "id": str(uuid4())})
            self.eval_specs[conf_key] = EvaluationSpecification(
                configuration=conf, settings={}, optimizer_info=opt_info
            )
            self.evaluation_data[self.current_stage].append(Datum(conf_key, "QUEUED"))
            # To understand recursion, you first must understand recursion :)
            return self.generate_evaluation_specification()

        # at this point there are pending evaluations and this iteration has to wait
        return None

    def digest_evaluation(
        self, evaluation_specificiation_id: UUID, evaluation: Evaluation
    ):
        """Registers the result of an evaluation.

        Args:
            evaluation_specificiation_id: [description]
            evaluation: [description]
        """
        self.config_sampler.digest_evaluation(evaluation)
        i = self.pending_evaluations.pop(evaluation_specificiation_id)
        d = self.evaluation_data[self.current_stage][i]
        d.status = "FINISHED" if not evaluation.all_objectives_none else "CRASHED"
        objective_value = evaluation.objectives[self.objective.name]
        if objective_value is not None:
            d.loss = (
                -objective_value
                if self.objective.greater_is_better
                else objective_value
            )

        # quick check if all configurations have finished yet
        if len(self.evaluation_data[self.current_stage]) == self.num_configs[
            self.current_stage
        ] and all(
            [
                e.status in ["FINISHED", "CRASHED"]
                for e in self.evaluation_data[self.current_stage]
            ]
        ):
            self._progress_to_next_stage()

    def _progress_to_next_stage(self):
        """Implements logic to promote configurations to the next stage."""
        # filter out crashed configurations
        data = [
            d for d in self.evaluation_data[self.current_stage] if np.isfinite(d.loss)
        ]
        self.current_stage += 1
        if self.current_stage == len(self.num_configs):
            self.finished = True
            return

        config_keys = self.config_promotion_function(
            data, self.num_configs[self.current_stage]
        )
        self.logger.debug(
            "Iteration %i: Advancing configurations %s to stage %i.",
            self.iteration,
            str(config_keys),
            self.current_stage,
        )
        self.evaluation_data.append(
            [Datum(config_key, "QUEUED") for config_key in config_keys]
        )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

from typing import List, Optional

import numpy as np

from blackboxopt.base import Objective
from blackboxopt.evaluation import Evaluation
from blackboxopt.optimizers.staged.iteration import Datum


def greedy_promotion(data: List[Datum], num_configs: int) -> list:
    """Promotes the best configurations to the next stage solely relying on the current
    loss.

    Args:
        data: List with all successful evaluations for this stage. All failed
            configurations have already been removed.
        num_configs: Maximum number of configurations to be promoted.

    Returns:
        List of the config_keys to be evaluated on the next higher fidelity. These must
        only include config_keys found in `data` und must also be of at most length
        `num_configs`. If fewer ids are returned, the remaining configurations for the
        next stage will be sampled using the `config_sample_function` of the
        staged_iteration.
    """
    losses = [d.loss for d in data]
    ranks = np.argsort(np.argsort(losses))
    n = min(num_configs, len(data))
    return [datum.config_key for rank, datum in zip(ranks, data) if rank < n]


def best_evaluation_at_highest_fidelity(
    evaluations: List[Evaluation],
    objective: Objective,
) -> Optional[Evaluation]:
    """From given list of evaluations, get the best in terms of minimal loss at the
    highest fidelity.

    Args:
        evaluations: [description]
        objective: [description]

    Returns:
        [description]
    """
    if not evaluations:
        return None

    successes = [
        evaluation
        for evaluation in evaluations
        if evaluation.objectives[objective.name] is not None
    ]
    if not successes:
        return None

    successful_fidelities = [
        evaluation.settings["fidelity"] for evaluation in successes
    ]
    if not successful_fidelities:
        return None

    max_successful_fidelities = max(successful_fidelities)
    successful_max_fidelity_evaluations = [
        evaluation
        for evaluation in successes
        if evaluation.settings["fidelity"] == max_successful_fidelities
    ]

    if not successful_max_fidelity_evaluations:
        return None

    sort_function = max if objective.greater_is_better else min
    best_evaluation = sort_function(
        successful_max_fidelity_evaluations, key=lambda e: e.objectives[objective.name]
    )

    return best_evaluation



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import abc
import logging
from typing import Dict, Iterable, List, Union

from parameterspace import ParameterSpace

from blackboxopt import (
    Evaluation,
    EvaluationSpecification,
    Objective,
    OptimizationComplete,
    OptimizerNotReady,
)
from blackboxopt.base import (
    SingleObjectiveOptimizer,
    call_functions_with_evaluations_and_collect_errors,
)
from blackboxopt.optimizers.staged.iteration import StagedIteration


def _validate_optimizer_info_id(evaluation: Evaluation):
    if evaluation.optimizer_info.get("id") is None:
        raise ValueError("Optimizer info is missing id.")


class StagedIterationOptimizer(SingleObjectiveOptimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objective: Objective,
        num_iterations: int,
        seed: int = None,
        logger: logging.Logger = None,
    ):
        """Base class for optimizers using iterations that compare configurations at
        different fidelities and race them in stages, like Hyperband or BOHB.

        Args:
            search_space: [description]
            objective: [description]
            num_iterations: The number of iterations that the optimizer will run.
            seed: [description]
            logger: [description]
        """
        super().__init__(search_space=search_space, objective=objective, seed=seed)
        self.logger = logging.getLogger("blackboxopt") if logger is None else logger
        self.num_iterations = num_iterations
        self.iterations: List[StagedIteration] = []
        self.evaluation_uuid_to_iteration: Dict[str, int] = {}
        self.pending_configurations: Dict[str, EvaluationSpecification] = {}

    def report(self, evaluations: Union[Evaluation, Iterable[Evaluation]]) -> None:
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations

        call_functions_with_evaluations_and_collect_errors(
            [super().report, _validate_optimizer_info_id, self._report],
            _evals,
        )

    def _report(self, evaluation: Evaluation) -> None:
        evaluation_specification_id = evaluation.optimizer_info.get("id")
        self.pending_configurations.pop(str(evaluation_specification_id))
        idx = self.evaluation_uuid_to_iteration.pop(str(evaluation_specification_id))
        self.iterations[idx].digest_evaluation(evaluation_specification_id, evaluation)

    def generate_evaluation_specification(self) -> EvaluationSpecification:
        """Get next configuration and settings to evaluate.

        Raises:
            OptimizationComplete: When the optimization run is finished, e.g. when the
                budget has been exhausted.
            OptimizerNotReady: When the optimizer is not ready to propose a new
                evaluation specification.
        """
        # check if any of the already active iterations returns a configuration and
        # simply return that
        for idx, iteration in enumerate(self.iterations):
            es = iteration.generate_evaluation_specification()

            if es is not None:
                self.evaluation_uuid_to_iteration[str(es.optimizer_info["id"])] = idx
                self.pending_configurations[str(es.optimizer_info["id"])] = es
                return es

        # if that didn't work, check if there another iteration can be started and then
        # ask it for a configuration
        if len(self.iterations) < self.num_iterations:
            self.iterations.append(self._create_new_iteration(len(self.iterations)))
            es = self.iterations[-1].generate_evaluation_specification()
            self.evaluation_uuid_to_iteration[str(es.optimizer_info["id"])] = (
                len(self.iterations) - 1
            )
            self.pending_configurations[str(es.optimizer_info["id"])] = es
            return es

        # check if the optimization is already complete or whether the optimizer is
        # waiting for evaluation results -> raise corresponding error
        if all([iteration.finished for iteration in self.iterations]):
            raise OptimizationComplete

        raise OptimizerNotReady

    @abc.abstractmethod
    def _create_new_iteration(self, iteration_index):
        """Optimizer specific way to create a new
        `blackboxopt.optimizer.utils.staged_iteration.StagedIteration` object
        """



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import datetime
import itertools
import warnings
from collections import Counter
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io._html
import scipy.stats as sps

from blackboxopt import Evaluation, Objective
from blackboxopt.utils import get_loss_vector
from blackboxopt.visualizations import utils

QUALITATIVE_COLORS = px.colors.qualitative.G10

plotly.io._html.to_html = utils.patch_plotly_io_to_html(plotly.io._html.to_html)


class NoSuccessfulEvaluationsError(ValueError):
    pass


def evaluations_to_df(evaluations: List[Evaluation]) -> pd.DataFrame:
    """Convert evaluations into multi index dataframe.

    The evaluations will be casted to dictionaries which will be normalized.
    The keys of the dicts will be used as secondary column index. Evaluations
    with one or more missing objective-value will be dropped.

    Example:

    ```
    Evaluation(objectives={'loss_1': 1.0, 'loss_2': -0.0}, stacktrace=None, ...)
    ```

    Will be transformed into:

    |    objectives   | stacktrace | ... |  <- "group" index
    | loss_1 | loss_2 | stacktrace | ... |  <- "field" index
    | ------ | ------ | ---------- | --- |
    |    1.0 |   -0.0 | None       | ... |
    """
    if not evaluations or len(evaluations) == 0:
        raise NoSuccessfulEvaluationsError

    # Filter out e.g. EvaluationSpecifications which might be passed into
    evaluations = [e for e in evaluations if isinstance(e, Evaluation)]

    # Transform to dicts, filter out evaluations with missing objectives
    evaluation_dicts = [e.__dict__ for e in evaluations if not e.any_objective_none]

    if len(evaluation_dicts) == 0:
        raise NoSuccessfulEvaluationsError

    df = pd.DataFrame(evaluation_dicts)

    # Flatten json/dict columns into single multi-index dataframe
    dfs_expanded = []
    for column in df.columns:
        # Normalize json columns keep original column for non-json columns
        try:
            df_temp = pd.json_normalize(df[column], errors="ignore", max_level=0)
        except AttributeError:
            df_temp = df[[column]]

        # Use keys of dicts as second level of column index
        df_temp.columns = pd.MultiIndex.from_product(
            [[column], df_temp.columns], names=["group", "field"]
        )
        # Drop empty columns
        df_temp = df_temp.dropna(axis=1, how="all")

        dfs_expanded.append(df_temp)

    df = pd.concat(dfs_expanded, join="outer", axis=1)

    # Parse datetime columns
    date_columns = [c for c in df.columns if "unixtime" in str(c)]
    df[date_columns] = df[date_columns].apply(pd.to_datetime, unit="s")

    # Calculate duration in seconds
    df["duration", "duration"] = (
        df["finished_unixtime", "finished_unixtime"]
        - df["created_unixtime", "created_unixtime"]
    )

    return df


def create_hover_information(sections: dict) -> Tuple[str, List]:
    """
    Create a [hovertemplate](https://plotly.com/python/reference/pie/#pie-hovertemplate)
    which is used to render hover hints in plotly charts.

    The data for the chart hovertext has to be provided as `custom_data` attribute to
    the chart and can be e.g. a list of column names.

    One oddness is, that in the template the columns can't be referenced by name, but
    only by index. That's why it is important to have the same ordering in the template
    as in the `custom_data` and the reason why this is done together in one function.

    Args:
        sections: Sections to render. The kyeys will show up as the section titles,
            values are expected to be a list of column names to be rendered under
            the section. E.g.: { "info": ["Objective #1", "Objective #2", "fidelity"] }

    Returns:
        (plotly hover template, data column names)
    """
    template = ""
    idx = 0
    for section, columns in sections.items():
        template += f"<br><b>{section.replace('_', ' ').title()}</b><br>"
        for column in columns:
            template += f"{column}: %{{customdata[{idx}]}}<br>"
            idx += 1
    template += "<extra></extra>"

    data_columns: list = sum(sections.values(), [])

    return template, data_columns


def _prepare_for_multi_objective_visualization(
    df: pd.DataFrame, objectives: Sequence[Objective]
) -> Tuple[pd.DataFrame, Dict[str, List[str]]]:
    # Unify potentially different objective directions according to greater_is_better
    loss_matrix = np.array(
        [
            get_loss_vector(known_objectives=objectives, reported_objectives=ovs)
            for ovs in df["objectives"].to_dict("records")
        ]
    )
    df["pareto efficient", "pareto efficient"] = utils.mask_pareto_efficient(
        loss_matrix
    )

    # Map column names (from field index) with "sections" shown in the mouse-over text
    # TODO: Handle case, where e.g. a column in user_info produces naming collision with
    #       a column from another group-index
    hover_sections: Dict[str, list] = {
        "info": [o.name for o in objectives]
        + ["pareto efficient", "created_unixtime", "duration"]
    }
    for section in ["configuration", "user_info", "settings", "optimizer_info"]:
        if section in df.columns:
            hover_sections[section] = list(df[section].columns)

    # Drop group index for visualizing in plotly
    df.columns = df.columns.droplevel("group")

    # Formatting for nicer output
    df["duration"] = df["duration"].round("s").astype("str")
    df["created_unixtime"] = df["created_unixtime"].dt.round("1s")

    return df, hover_sections


def multi_objective_visualization(
    evaluations: List[Evaluation], objectives: Sequence[Objective]
):
    if not evaluations:
        raise NoSuccessfulEvaluationsError

    # Prepare dataframe for visualization
    df = evaluations_to_df(evaluations)
    df, hover_sections = _prepare_for_multi_objective_visualization(df, objectives)

    # Create hover template and list of corresponding dataframe column names
    hover_template, hover_data_columns = create_hover_information(hover_sections)

    fig_kwargs = dict(
        data_frame=df,
        color="pareto efficient",
        color_discrete_map={
            False: QUALITATIVE_COLORS[5],
            True: QUALITATIVE_COLORS[2],
        },
        custom_data=hover_data_columns,
        title="[BBO] Multi-objective pairwise scatter plot",
    )
    traces_kwargs: Dict[str, Union[str, dict, bool]] = dict(
        hovertemplate=hover_template,
        marker=dict(line=dict(width=1, color="white")),
    )

    if len(objectives) == 2:
        fig = px.scatter(**fig_kwargs, x=objectives[0].name, y=objectives[1].name)
    else:
        fig = px.scatter_matrix(**fig_kwargs, dimensions=[o.name for o in objectives])
        traces_kwargs["diagonal_visible"] = False

    fig.update_traces(**traces_kwargs)
    return fig


def parallel_coordinate_plot_parameters(
    evaluations: List[Evaluation],
    columns: Optional[List[str]] = None,
    color_by: Optional[str] = None,
):
    """Create an interactive parallel coordinate plot.

    Useful to investigate relationships in a higher dimensional search space and the
        optimization's objective(s).

    Args:
        evaluations: Evaluations to plot.
        columns: Names of columns to show. Can contain parameter names, objective names
            and settings keys. If `None`, all parameters, objectives and settings are
            displayed.
        color_by: Parameter name, objective name or settings key. The corresponding
            column will be shown at the very right, it's value will be used for the
            color scale. If `None`, all lines have the same color.

    Returns:
        Plotly figure

    Raised:
        NoSuccessfulEvaluationsError: In case `evaluations` does not contain at least
            one successful evaluation (an evaluation with objective value != `None`).
    """
    if not evaluations:
        raise NoSuccessfulEvaluationsError

    # Prepare dataframe for visualization
    df = evaluations_to_df(evaluations)

    # Drop unused columns and indices
    if "settings" in df.columns:
        df = df[["configuration", "settings", "objectives"]]
        settings_cols = df["settings"].columns.to_list()
    else:
        df = df[["configuration", "objectives"]]
        settings_cols = []
    objective_cols = df["objectives"].columns.to_list()
    df = df.droplevel(0, axis=1)

    # If no columns are specified, use all:
    if not columns:
        columns = df.columns.to_list()

    if color_by and color_by not in columns:
        raise ValueError(
            f"Unknown column name in color_by='{color_by}'. Please make sure, that this"
            + "column name is correct and one of the visible columns."
        )

    ambigious_columns = [k for k, v in Counter(df[columns].columns).items() if v > 1]
    if ambigious_columns:
        raise ValueError(
            "All columns to plot must have a unique name, but those are ambigious: "
            + f"{ambigious_columns}. Either rename parameters/settings/objective to "
            + "be unique or provide only the unambigious ones as `columns` argument."
        )

    # Prepare a coordinate (vertical line) for every column
    coordinates = []
    colored_coordinate = {}
    for column in columns:
        coordinate: Dict[str, Any] = {}

        if column in objective_cols:
            coordinate["label"] = f"<b>Objective: {column}</b>"
        elif column in settings_cols:
            coordinate["label"] = f"Setting: {column}"
        else:
            coordinate["label"] = column

        parameter_type = df[column].dtype.name
        if parameter_type.startswith("float") or parameter_type.startswith("int"):
            # Handling floats and integers the same, because unfortunately it's hard to
            # use integers only for ticks and still be robust regarding a large range
            # of values.
            coordinate["values"] = df[column]
        elif parameter_type in ["object", "bool"]:
            # Encode categorical values to integers. Unfortunately, ordinal parameters
            # loose there ordering, as there is no information about the order in the
            # evaluations.
            # The string conversion is necessary for unhashable entries, e.g. of
            # type List, which can't be casted to categories.
            df[column] = df[column].astype(str).astype("category")
            categories = df[column].cat.categories.to_list()
            encoded_categories = list(range(len(categories)))
            df[column].cat.categories = encoded_categories
            # Use integer encodings for scale and category values as tick labels
            coordinate["ticktext"] = categories
            coordinate["tickvals"] = encoded_categories
            coordinate["values"] = df[column].astype("str")
        else:
            warnings.warn(
                f"Ignoring column with unknown type: {column}<{parameter_type}>"
            )
            continue

        if column == color_by:
            colored_coordinate = coordinate
        else:
            coordinates.append(coordinate)

    # Append colored coordinate to the end (right)
    if colored_coordinate:
        coordinates.append(colored_coordinate)

    # Plot
    return go.Figure(
        data=go.Parcoords(
            line=dict(
                # Color lines by objective value
                color=df[color_by] if color_by else None,
                colorscale=px.colors.diverging.Tealrose,
                showscale=True,
                # Use colorbar as kind of colored extension to the axis
                colorbar=dict(
                    thickness=16, x=1, xpad=0, ypad=1, tickmode="array", tickvals=[]
                ),
            ),
            dimensions=coordinates,
        ),
        layout=dict(title="[BBO] Parallel coordinates plot"),
    )


class Visualizer:
    def __init__(self, evaluations: List[Evaluation], objective: Objective):
        times_finished = np.array(utils.get_times_finished(evaluations))
        if len(times_finished) == 0 or np.isnan(times_finished).all():
            raise NoSuccessfulEvaluationsError()

        self.objective = objective

        self.times = times_finished - utils.get_t0(evaluations)
        sort_idx = np.argsort(self.times)
        self.times = self.times[sort_idx]
        self.evaluations = [evaluations[i] for i in sort_idx]
        self.objective_values = np.array(
            utils.get_objective_values(self.evaluations), dtype=float
        )
        self.durations = np.array(utils.get_durations(self.evaluations))
        self.fidelities = np.array(utils.get_fidelities(self.evaluations))
        self.configs = utils.get_configs(self.evaluations)

        all_keys = [set(c.keys()) for c in self.configs]
        self.all_config_keys: List[str] = sorted(set().union(*all_keys))
        self.all_fidelities: List[float] = sorted(set(self.fidelities))

        tmp_dict: Dict["str", Union[np.ndarray, List[str]]] = {}
        for key in self.all_config_keys:
            tmp_dict[key] = np.array([c.get(key, "N/A") for c in self.configs])
        tmp_dict["fidelity"] = ["%3.2e" % f for f in self.fidelities]

        self.info_dicts = [
            {
                self.objective.name: "%3.2e" % self.objective_values[i],
                "duration": str(datetime.timedelta(seconds=int(self.durations[i])))
                if np.isfinite(self.durations[i])
                else "N/A",
                "fidelity": "%3.2e" % self.fidelities[i],
            }
            for i in range(self.objective_values.shape[0])
        ]

        self.optimizer_info_dicts = [e["optimizer_info"] for e in self.evaluations]

    def objective_over_time(self, x_range=None, y_range=None, log_x=False, log_y=False):
        colors = px.colors.qualitative.G10
        fig = go.Figure()

        for i, f in enumerate(self.all_fidelities):

            mask = self.fidelities == f

            hover_texts = utils.get_hover_texts(
                self.info_dicts, self.optimizer_info_dicts, self.configs, mask
            )

            fig.add_scatter(
                x=self.times[mask],
                y=self.objective_values[mask],
                name="%3.2e" % f,
                mode="markers",
                marker=dict(color=colors[i]),
                legendgroup=str(f),
                hovertemplate="%{text} <extra></extra>",
                text=hover_texts,
            )

            times, objectives = utils.get_incumbent_objective_over_time_single_fidelity(
                self.objective, self.objective_values, self.times, self.fidelities, f
            )
            fig.add_scatter(
                x=times,
                y=objectives,
                mode="lines",
                line=dict(color=colors[i]),
                legendgroup=str(f),
                showlegend=False,
                hoverinfo="none",
            )

        utils.add_plotly_buttons_for_logscale(fig)
        utils.plotly_set_axis(fig, x_range, y_range, log_x, log_y)
        fig.update_layout(
            title=f"[BBO] Reported {self.objective.name} over time",
            legend_title_text="Fidelity",
            xaxis_title="Time [s]",
            yaxis_title=self.objective.name,
        )

        return fig

    def objective_over_duration(
        self, x_range=None, y_range=None, log_x=False, log_y=False
    ):
        colors = px.colors.qualitative.G10
        fig = go.Figure()

        for i, f in enumerate(self.all_fidelities):

            mask = self.fidelities == f

            hover_texts = utils.get_hover_texts(
                self.info_dicts, self.optimizer_info_dicts, self.configs, mask
            )

            fig.add_scatter(
                x=self.durations[mask],
                y=self.objective_values[mask],
                name="%3.2e" % f,
                mode="markers",
                marker=dict(color=colors[i]),
                legendgroup=str(f),
                hovertemplate="%{text} <extra></extra>",
                text=hover_texts,
            )

        utils.add_plotly_buttons_for_logscale(fig)
        utils.plotly_set_axis(fig, x_range, y_range, log_x, log_y)
        fig.update_layout(
            title=f"[BBO] Reported {self.objective.name} over duration",
            legend_title_text="Fidelity",
            xaxis_title="Duration [s]",
            yaxis_title=self.objective.name,
        )

        return fig

    def cdf_objective_values(self, x_range=None, log_x=False):

        colors = px.colors.qualitative.G10
        fig = go.Figure()

        for i, f in enumerate(self.all_fidelities):

            mask = self.fidelities == f
            x, y = utils.get_cdf_x_and_y(self.objective_values[mask])

            fig.add_scatter(
                x=x,
                y=y,
                name="%3.2e" % f,
                mode="lines",
                marker=dict(color=colors[i]),
                legendgroup=str(f),
            )

        utils.add_plotly_buttons_for_logscale(fig)
        utils.plotly_set_axis(fig, x_range, None, log_x)
        fig.update_layout(
            title=f"[BBO] CDF of {self.objective.name} by fidelity",
            legend_title_text="Fidelity",
            xaxis_title=self.objective.name,
            yaxis_title="CDF",
        )

        return fig

    def cdf_durations(self, x_range=None, log_x=False):

        colors = px.colors.qualitative.G10
        fig = go.Figure()

        for i, f in enumerate(self.all_fidelities):

            mask = self.fidelities == f
            x, y = utils.get_cdf_x_and_y(self.durations[mask])

            fig.add_scatter(
                x=x,
                y=y,
                name="%3.2e" % f,
                mode="lines",
                marker=dict(color=colors[i]),
                legendgroup=str(f),
            )

        utils.add_plotly_buttons_for_logscale(fig)
        utils.plotly_set_axis(fig, x_range, None, log_x)
        fig.update_layout(
            title="[BBO] CDF of durations by fidelity",
            legend_title_text="Fidelity",
            xaxis_title="Duration [s]",
            yaxis_title="CDF",
        )

        return fig

    def correlation_coefficients(self):
        matrix = utils.get_objective_values_matrix(self.evaluations)
        fidelities = sorted(list(set(self.fidelities)))

        corr_coeffs = np.full([len(fidelities), len(fidelities)], np.nan)
        p_values = np.full([len(fidelities), len(fidelities)], np.nan)
        num_samples = np.full([len(fidelities), len(fidelities)], 0)
        for i, j in itertools.combinations(range(len(fidelities)), 2):

            row_mask = np.all(np.isfinite(matrix[:, [i, j]]), axis=1)
            v1, v2 = matrix[row_mask][:, [i, j]].T
            res = sps.spearmanr(v1, v2)
            corr_coeffs[i, j] = res.correlation
            p_values[i, j] = res.pvalue
            num_samples[i, j] = len(v1)

        fig = go.Figure()

        hovertext = [
            [
                "<b>Correlation for %.2e and %.2e</b><br />"
                % (fidelities[i], fidelities[j])
                + "ρ:  %.3f<br />" % (corr_coeffs[i, j])
                + "p value: %.2e<br />" % (p_values[i, j])
                + "N samples: %i" % num_samples[i, j]
                for i in range(len(fidelities) - 1)
            ]
            for j in range(1, len(fidelities))
        ]

        fig.add_trace(
            go.Heatmap(
                z=corr_coeffs[:-1, 1:].T,
                y=["%3.2e" % f for f in fidelities[1:]],
                x=["%3.2e" % f for f in fidelities[:-1]],
                text=hovertext,
                hovertemplate="%{text}<extra></extra>",
                colorscale="viridis",
            )
        )
        fig.update_xaxes(type="category")
        fig.update_yaxes(type="category")

        fig.update_layout(
            title=f"[BBO] {self.objective.name} rank correlation for configurations "
            + "across fidelities",
            xaxis_title="Fidelity",
            yaxis_title="Fidelity",
        )

        return fig



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import importlib.resources
from functools import wraps
from typing import Callable, List

import numpy as np

import blackboxopt
from blackboxopt import Evaluation, Objective

# For backwards compatibility; TODO: remove with next major release
from blackboxopt.utils import mask_pareto_efficient


def get_t0(evaluations: List[Evaluation]):
    return min([e.created_unixtime for e in evaluations])


def get_objective_values(evaluations: List[Evaluation]):
    return [
        list(e.objectives.values())[0] if not e.any_objective_none else float("NaN")
        for e in evaluations
    ]


def get_times_finished(evaluations: List[Evaluation]):
    return [
        e.finished_unixtime if not e.any_objective_none else float("NaN")
        for e in evaluations
    ]


def get_durations(evaluations: List[Evaluation]):
    return [
        e.finished_unixtime - e.created_unixtime
        if not e.any_objective_none
        else float("NaN")
        for e in evaluations
    ]


def get_fidelities(evaluations: List[Evaluation]):
    return [e.settings.get("fidelity", -1.0) for e in evaluations]


def get_bohb_ids(evaluations: List[Evaluation]):
    return [e.optimizer_info["configuration_key"] for e in evaluations]


def get_configs(evaluations: List[Evaluation]):
    return [e.configuration for e in evaluations]


def get_objective_values_matrix(evaluations: List[Evaluation]):

    # need to turn config dicts into tuples to use set
    config_tuples = {tuple(e.configuration.values()) for e in evaluations}
    configs = list(config_tuples)

    fidelities = sorted(set(get_fidelities(evaluations)))

    matrix = np.full([len(configs), len(fidelities)], np.nan)

    for e in evaluations:
        row = configs.index(tuple(e.configuration.values()))
        column = fidelities.index(e.settings["fidelity"])
        try:
            matrix[row, column] = list(e.objectives.values())[0]
        except Exception:
            continue

    return matrix


def get_incumbent_objective_over_time_single_fidelity(
    objective: Objective,
    objective_values: np.ndarray,
    times: np.ndarray,
    fidelities: np.ndarray,
    target_fidelity: float,
):
    """Filter for results with given target fidelity and generate incumbent trace."""
    # filter out fidelity and take min/max of objective_values
    idx = np.logical_and(fidelities == target_fidelity, np.isfinite(objective_values))
    _times = times[idx]
    if objective.greater_is_better:
        _objective_values = np.maximum.accumulate(objective_values[idx])
    else:
        _objective_values = np.minimum.accumulate(objective_values[idx])
    # get unique objective values and sort their indices (to be in chronological order)
    _, idx = np.unique(_objective_values, return_index=True)
    idx.sort()
    # find objective_values
    _objective_values = _objective_values[idx]
    _times = _times[idx]
    # add steps where a new incumbent was found
    _times = np.repeat(_times, 2)[1:]
    _objective_values = np.repeat(_objective_values, 2)[:-1]
    # append best value for largest time to extend the lines
    _times = np.concatenate([_times, np.nanmax(times, keepdims=True)])
    _objective_values = np.concatenate([_objective_values, _objective_values[-1:]])
    return _times, _objective_values


def dict_to_hovertext(d):
    strings = [f"{key}: {d[key]}" for key in sorted(d.keys())]
    return "<br />".join(strings)


def get_hover_texts(info_dicts, optimizer_info_dicts, config_dicts, mask):
    info_texts = [dict_to_hovertext(info_dicts[j]) for j, k in enumerate(mask) if k]
    optimizer_texts = [
        dict_to_hovertext(optimizer_info_dicts[j]) for j, k in enumerate(mask) if k
    ]
    config_texts = [dict_to_hovertext(config_dicts[j]) for j, k in enumerate(mask) if k]
    hover_texts = [
        f"<b>Info</b><br />{i}<br /><br />"
        + f"<b>Optimizer Info</b><br />{o}<br /><br />"
        + f"<b>Configuration</b><br />{c}"
        for i, o, c in zip(info_texts, optimizer_texts, config_texts)
    ]
    return hover_texts


def get_cdf_x_and_y(values):
    values = np.sort(values)
    y = np.linspace(0, 1, len(values), endpoint=True)
    return values, y


def plotly_set_axis(fig, x_range=None, y_range=None, log_x=False, log_y=False):
    if x_range is not None:
        fig.update_xaxes(range=x_range)
    if y_range is not None:
        fig.update_yaxes(range=y_range)

    if log_x:
        fig.update_xaxes(type="log")
    if log_y:
        fig.update_yaxes(type="log")


def add_plotly_buttons_for_logscale(fig):
    updatemenus = [
        dict(
            type="buttons",
            direction="down",
            buttons=list(
                [
                    dict(
                        args=[{"yaxis.type": "linear"}],
                        label="Linear y",
                        method="relayout",
                    ),
                    dict(
                        args=[{"yaxis.type": "log"}], label="Log y", method="relayout"
                    ),
                    dict(
                        args=[{"xaxis.type": "linear"}],
                        label="Linear x",
                        method="relayout",
                    ),
                    dict(
                        args=[{"xaxis.type": "log"}], label="Log x", method="relayout"
                    ),
                ]
            ),
        ),
    ]

    fig.update_layout(updatemenus=updatemenus)


def patch_plotly_io_to_html(method: Callable) -> Callable:
    """Patch `plotly.io.to_html` with additional javascript to improve usability.

    Might become obsolete, when https://github.com/plotly/plotly.js/issues/998 gets
    fixed.

    Injects `<script>`-tag with content from `to_html_patch.js` at the end of the HTML
    output. But only, if the chart title starts with "[BBO]" (to minimize side
    effects, if the user uses `plotly.io` for something else).

    `plotly.io.to_html` is also internally used for `figure.show()` and
    `figure.to_html()`, so this is covered, too.

    Args:
        method: Original `plotly.io.to_html` method.

    Returns:
        Patched method.
    """

    @wraps(method)
    def wrapped(*args, **kwargs):
        html = method(*args, **kwargs)

        # Test if title text contains "[BBO]"
        if html.find('"title":{"text":"[BBO]') < 0:
            return html

        js = importlib.resources.read_text(
            blackboxopt.visualizations, "to_html_patch.js"
        )
        html_to_inject = f"<script>{js}</script>"
        insert_idx = html.rfind("</body>")
        if insert_idx >= 0:
            # Full html page got rendered, inject <script> before <\body>
            html = html[:insert_idx] + html_to_inject + html[insert_idx:]
        else:
            # Only chart part got rendered: append <script> at the end
            html = html + html_to_inject

        return html

    return wrapped



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import functools
import logging
import warnings
from typing import Callable, Dict, Iterable, Optional, Tuple, Union

from blackboxopt import (
    Evaluation,
    EvaluationSpecification,
    Objective,
    OptimizerNotReady,
    sort_evaluations,
)
from blackboxopt.base import (
    SingleObjectiveOptimizer,
    call_functions_with_evaluations_and_collect_errors,
    validate_objectives,
)

try:
    import numpy as np
    import parameterspace as ps
    import scipy.optimize as sci_opt
    import torch
    from botorch.acquisition import AcquisitionFunction
    from botorch.models.model import Model
    from botorch.optim import optimize_acqf
    from botorch.sampling.samplers import IIDNormalSampler
    from sklearn.impute import SimpleImputer

except ImportError as e:
    raise ImportError(
        "Unable to import BOTorch optimizer specific dependencies. "
        + "Make sure to install blackboxopt[botorch]"
    ) from e


def impute_nans_with_constant(x: torch.Tensor, c: float = -1.0) -> torch.Tensor:
    """Impute `NaN` values with given constant value.

    Args:
        x: Input tensor of shape `n x d` or `b x n x d`.
        c: Constant used as fill value to replace `NaNs`.

    Returns:
        - x_i - `x` where all `NaN`s are replaced with given constant.
    """
    if x.numel() == 0:  # empty tensor, nothing to impute
        return x
    x_i = x.clone()

    # cast n x d to 1 x n x d (cover non-batch case)
    if len(x.shape) == 2:
        x_i = x_i.reshape(torch.Size((1,)) + x_i.shape)

    for b in range(x_i.shape[0]):
        x_1 = x_i[b, :, :]
        x_1 = torch.tensor(
            SimpleImputer(
                missing_values=np.nan, strategy="constant", fill_value=c
            ).fit_transform(x_1),
            dtype=x.dtype,
        )
        x_i[b, :, :] = x_1

    # cast 1 x n x d back to n x d if originally non-batch
    if len(x.shape) == 2:
        x_i = x_i.reshape(x.shape)
    return x_i


def to_numerical(
    evaluations: Iterable[Evaluation],
    search_space: ps.ParameterSpace,
    objective: Objective,
    batch_shape: torch.Size = torch.Size(),
    torch_dtype: torch.dtype = torch.float32,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Convert evaluations to one `(#batch, #evaluations, #parameters)` tensor
    containing the numerical representations of the configurations and
    one `(#batch, #evaluations, 1)` tensor containing the loss representation of
    the evaluations' objective value (flips the sign for objective value
    if `objective.greater_is_better=True`).

    Args:
        evaluations: List of evaluations that were collected during optimization.
        search_space: Search space used during optimization.
        objective: Objective that was used for optimization.
        batch_shape: Batch dimension(s) used for batched models.
        torch_dtype: Type of returned tensors.

    Returns:
        - X: Numerical representation of the configurations
        - Y: Numerical representation of the objective values

    Raises:
        ValueError: If one of configurations is not valid w.r.t. search space.
        ValueError: If one of configurations includes parameters that are not part of
            the search space.
    """
    # validate configuration values and dimensions
    parameter_names = search_space.get_parameter_names() + list(
        search_space.get_constant_names()
    )
    for e in evaluations:
        with warnings.catch_warnings():
            # we already raise error if search space not valid, thus can ignore warnings
            warnings.filterwarnings(
                "ignore", category=RuntimeWarning, message="Parameter"
            )
            if not search_space.check_validity(e.configuration):
                raise ValueError(
                    f"The provided configuration {e.configuration} is not valid."
                )
        if not set(parameter_names) >= set(e.configuration.keys()):
            raise ValueError(
                f"Mismatch in parameter names from search space {parameter_names} and "
                + f"configuration {e.configuration}"
            )

    X = torch.tensor(
        np.array([search_space.to_numerical(e.configuration) for e in evaluations]),
        dtype=torch_dtype,
    )
    X = X.reshape(*batch_shape + X.shape)
    Y = torch.tensor(
        np.array([[e.objectives[objective.name]] for e in evaluations], dtype=float),
        dtype=torch_dtype,
    )
    Y = Y.reshape(*batch_shape + Y.shape)

    if objective.greater_is_better:
        Y *= -1

    return X, Y


def init_af_opt_kwargs(af_opt_kwargs: Optional[dict]) -> dict:
    """Provide default initialization for the acquisition function optimizer
    configuration. Ensure that all mandatory fields are set to be used by BoTorch.

    Args:
        af_opt_kwargs: Acquisition function configuration.

    Returns:
        Acquisition function optimizer configuration, applicable for BoTorch's
            optimizer.
    """

    af_opt_config = {} if af_opt_kwargs is None else af_opt_kwargs

    # number of initial samples during AF optimization
    af_opt_config.setdefault("raw_samples", 1024)

    # number of restarts during AF optimization
    af_opt_config.setdefault("num_restarts", 4)

    return af_opt_config


class SingleObjectiveBOTorchOptimizer(SingleObjectiveOptimizer):
    def __init__(
        self,
        search_space: ps.ParameterSpace,
        objective: Objective,
        model: Model,
        acquisition_function_factory: Callable[[Model], AcquisitionFunction],
        af_optimizer_kwargs=None,
        num_initial_random_samples: int = 1,
        max_pending_evaluations: Optional[int] = 1,
        batch_shape: torch.Size = torch.Size(),
        logger: Optional[logging.Logger] = None,
        seed: Optional[int] = None,
        torch_dtype: torch.dtype = torch.float64,
    ):
        """Single objective BO optimizer that uses as a surrogate model the `model`
        object provided by user.

        The `model` is expected to be extended from BoTorch base model `Model` class,
        and does not require to be a GP model.

        Args:
            search_space: The space in which to optimize.
            objective: The objective to optimize.
            model: Surrogate model of `Model` type.
            acquisition_function_factory: Callable that produces an acquisition function
                instance, could also be a compatible acquisition function class.
                Only acquisition functions to be minimized are supported.
                Providing a partially initialized class is possible with, e.g.
                `functools.partial(UpperConfidenceBound, beta=6.0, maximize=False)`.
            af_optimizer_kwargs: Settings for acquisition function optimizer,
                see `botorch.optim.optimize_acqf`.
            num_initial_random_samples: Size of the initial space-filling design that
                is used before starting BO. The points are sampled randomly in the
                search space. If no random sampling is required, set it to 0.
            max_pending_evaluations: Maximum number of parallel evaluations. For
                sequential BO use the default value of 1. If no limit is required,
                set it to None.
            batch_shape: Batch dimension(s) used for batched models.
            logger: Custom logger.
            seed: A seed to make the optimization reproducible.
            torch_dtype: Torch data type used for storing the data. This needs to match
                the dtype of the model used
        """
        super().__init__(search_space=search_space, objective=objective, seed=seed)
        self.num_initial_random = num_initial_random_samples
        self.max_pending_evaluations = max_pending_evaluations
        self.batch_shape = batch_shape
        self.logger = logger or logging.getLogger("blackboxopt")

        self.torch_dtype = torch_dtype
        self.X = torch.empty(
            (*self.batch_shape, 0, len(search_space)), dtype=torch_dtype
        )
        self.losses = torch.empty((*self.batch_shape, 0, 1), dtype=torch_dtype)
        self.pending_specifications: Dict[int, EvaluationSpecification] = {}
        if seed is not None:
            torch.manual_seed(seed=seed)

        self.model = model
        self.acquisition_function_factory = acquisition_function_factory
        self.af_opt_kwargs = init_af_opt_kwargs(af_optimizer_kwargs)

    def _create_fantasy_model(self, model: Model) -> Model:
        """Create model with the pending specifications and model based
        outcomes added to the training data."""
        if not self.pending_specifications:
            # nothing to do when there are no pending specs
            return model

        pending_X = torch.tensor(
            np.array(
                [
                    self.search_space.to_numerical(e.configuration)
                    for e in self.pending_specifications.values()
                ]
            ),
            dtype=self.torch_dtype,
        )

        model = model.fantasize(pending_X, IIDNormalSampler(1), observation_noise=False)

        # model.fantasize extends model's X and Y with batch_size, even if originally
        # not given -> need to reshape these to their original representation
        n_samples = model.train_targets.size(-1)
        n_features = len(self.search_space)
        model.train_inputs[0] = model.train_inputs[0].reshape(
            torch.Size((*self.batch_shape, n_samples, n_features))
        )
        model.train_targets = model.train_targets.reshape(
            torch.Size((*self.batch_shape, n_samples, 1))
        )
        return model

    def _generate_evaluation_specification(self):
        """Optimize acquisition on fantasy model to pick next point."""
        fantasy_model = self._create_fantasy_model(self.model)
        fantasy_model.eval()

        # find next configuration by optimizing the acquisition function
        af = self.acquisition_function_factory(fantasy_model)
        if getattr(af, "maximize", False):
            raise ValueError(
                "Only acquisition functions that need to be minimized are supported. "
                f"The given {af.__class__.__name__} has maximize=True. "
                "One potential fix is using functools.partial("
                f"{af.__class__.__name__}, maximize=False) as the "
                "acquisition_function_factory init argument."
            )

        # numerical representation always lives on hypercube
        bounds = torch.tensor([[0, 1]] * len(self.search_space)).float().T

        configuration, _ = optimize_acqf(
            af,
            bounds=bounds,
            q=1,
            **self.af_opt_kwargs,
        )

        return EvaluationSpecification(
            configuration=self.search_space.from_numerical(configuration[0]),
        )

    def generate_evaluation_specification(self) -> EvaluationSpecification:
        """Call the optimizer specific function and append a unique integer id
        to the specification.

        Please refer to the docstring of
        `blackboxopt.base.SingleObjectiveOptimizer.generate_evaluation_specification`
        for a description of the method.
        """
        if (
            self.max_pending_evaluations
            and len(self.pending_specifications) == self.max_pending_evaluations
        ):
            raise OptimizerNotReady

        if self.num_initial_random > 0 and (
            self.X.size(-2) < self.num_initial_random
            or torch.nonzero(~torch.any(self.losses.isnan(), dim=1)).numel() == 0
        ):
            # We keep generating random samples until there are enough samples, and
            # at least one of them has a valid objective
            eval_spec = EvaluationSpecification(
                configuration=self.search_space.sample(),
            )
        else:
            eval_spec = self._generate_evaluation_specification()

        eval_id = self.X.size(-2) + len(self.pending_specifications)
        eval_spec.optimizer_info["evaluation_id"] = eval_id
        self.pending_specifications[eval_id] = eval_spec
        return eval_spec

    def _remove_pending_specifications(
        self, evaluations: Union[Evaluation, Iterable[Evaluation]]
    ):
        """Find and remove the corresponding entries in `self.pending_specifications`.

        Args:
            evaluations: List of completed evaluations.
        Raises:
            ValueError: If an evaluation is reported with an ID that was not issued
            by the optimizer, the method will fail.
        """
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations

        for e in _evals:
            if "evaluation_id" not in e.optimizer_info:
                self.logger.debug("User provided EvaluationSpecification received.")
                continue

            if e.optimizer_info["evaluation_id"] not in self.pending_specifications:
                msg = (
                    "Unknown evaluation_id reported. This could indicate that the "
                    "evaluation has been reported before!"
                )
                self.logger.error(msg)
                raise ValueError(msg)

            del self.pending_specifications[e.optimizer_info["evaluation_id"]]

    def _append_evaluations_to_data(
        self, evaluations: Union[Evaluation, Iterable[Evaluation]]
    ):
        """Convert the reported evaluation into its numerical representation
        and append it to the training data.

        Args:
            evaluations: List of completed evaluations.
        """
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations

        X, Y = to_numerical(
            _evals,
            self.search_space,
            self.objective,
            self.batch_shape,
            self.torch_dtype,
        )

        # fill in NaNs originating from inactive parameters (conditional spaces support)
        # botorch expect numerical representation of inputs to be within the unit
        # hypercube, thus we can't use the default c=-1.0
        X = impute_nans_with_constant(X, c=0.0)

        self.logger.debug(f"Next training configuration(s):{X}, {Y}")

        self.X = torch.cat([self.X, X], dim=-2)
        self.losses = torch.cat([self.losses, Y], dim=-2)

    def _update_internal_evaluation_data(
        self, evaluations: Iterable[Evaluation]
    ) -> None:
        """Check validity of the evaluations and do optimizer agnostic bookkeeping."""
        call_functions_with_evaluations_and_collect_errors(
            [
                functools.partial(validate_objectives, objectives=[self.objective]),
                self._remove_pending_specifications,
                self._append_evaluations_to_data,
            ],
            sort_evaluations(evaluations),
        )

    def report(self, evaluations: Union[Evaluation, Iterable[Evaluation]]) -> None:
        """A simple report method that conditions the model on data.
        This likely needs to be overridden for more specific BO implementations.
        """
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations
        self._update_internal_evaluation_data(_evals)
        # Just for populating all relevant caches
        self.model.posterior(self.X)
        # The actual model update
        self.model.condition_on_observations(self.X, self.losses)

    def predict_model_based_best(self) -> Optional[Evaluation]:
        """Get the current configuration that is estimated to be the best (in terms of
        optimal objective value) without waiting for a reported evaluation of that
        configuration. Instead, the objective value estimation relies on BO's
        underlying model.

        This might return `None` in case there is no successfully evaluated
        configuration yet (thus, the optimizer has not been given training data yet).

        Returns:
            blackboxopt.evaluation.Evaluation
                The evaluated specification containing the estimated best configuration
                or `None` in case no evaluations have been reported yet.
        """
        if self.model.train_inputs[0].numel() == 0:
            return None

        def posterior_mean(x):
            # function to be optimized: posterior mean
            # scipy's minimize expects the following interface:
            #  - input: 1-D array with shape (n,)
            #  - output: float
            mean = self.model.posterior(torch.from_numpy(np.atleast_2d(x))).mean
            return mean.item()

        # prepare initial random samples and bounds for scipy's minimize
        n_init_samples = 10
        init_points = np.asarray(
            [
                self.search_space.to_numerical(self.search_space.sample())
                for _ in range(n_init_samples)
            ]
        )
        bounds = self.search_space.get_continuous_bounds()

        # use scipy's minimize to find optimum of the posterior mean
        optimized_points = [
            sci_opt.minimize(
                fun=posterior_mean,
                constraints=None,
                jac=False,
                x0=x,
                args=(),
                bounds=bounds,
                method="L-BFGS-B",
                options=None,
            )
            for x in init_points
        ]

        f_optimized = np.array(
            [np.atleast_1d(p.fun) for p in optimized_points]
        ).flatten()
        # get indexes of optimum value (with a tolerance)
        inds = np.argwhere(np.isclose(f_optimized, np.min(f_optimized)))
        # randomly select one index if there are multiple
        ind = np.random.choice(inds.flatten())

        # create Evaluation from the best estimated configuration
        best_x = optimized_points[ind].x
        best_y = posterior_mean(best_x)
        return Evaluation(
            configuration=self.search_space.from_numerical(best_x),
            objectives={
                self.objective.name: -1 * best_y
                if self.objective.greater_is_better
                else best_y
            },
        )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import abc
import collections
import functools
from dataclasses import dataclass
from typing import Callable, Iterable, List, Tuple, Type, Union

from parameterspace import ParameterSpace

from blackboxopt.evaluation import Evaluation, EvaluationSpecification


class OptimizationComplete(Exception):
    """Exception that is raised when the optimization run is finished, e.g. when the
    budget has been exhausted.
    """


class OptimizerNotReady(Exception):
    """Exception that is raised when the optimizer is not ready to propose a new
    evaluation specification.
    """


class ObjectivesError(ValueError):
    """Raised on incomplete or missing objectives."""


class ConstraintsError(ValueError):
    """Raised on incomplete or missing constraints."""


class ContextError(ValueError):
    """Raised on incomplete or missing context information."""


class EvaluationsError(ValueError):
    """Raised on invalid evaluations.

    The problematic evaluations and their respective exceptions are passed in the
    `evaluations_with_errors` attribute.
    """

    def __init__(self, evaluations_with_errors: List[Tuple[Evaluation, Exception]]):
        self.message = (
            "An error with one or more evaluations occurred. Check the "
            "'evaluations_with_errors' attribute of this exception for details."
        )
        self.evaluations_with_errors = evaluations_with_errors


@dataclass
class Objective:
    name: str
    greater_is_better: bool


def _raise_on_duplicate_objective_names(objectives: List[Objective]) -> None:
    names = [o.name for o in objectives]
    duplications = [
        (name, count) for name, count in collections.Counter(names).items() if count > 1
    ]
    if duplications:
        raise ValueError(
            "All objectives need to have a unique name, but found duplications "
            + f"(objective name, occurrence count): {duplications}"
        )


def raise_on_unknown_or_incomplete(
    exception: Type[ValueError], known: Iterable[str], reported: Iterable[str]
) -> None:
    """Raise the given exception if not all known strings are contained in reported or
    the other way around.
    """
    known_set = set(known)
    reported_set = set(reported)

    unknown = reported_set - known_set
    if unknown:
        raise exception(
            f"Unknown reported: {list(unknown)}. Valid are only: {list(known_set)}"
        )
    missing = known_set - reported_set
    if missing:
        raise exception(f"Missing: {list(missing)}")


def validate_objectives(evaluation: Evaluation, objectives: List[Objective]):
    raise_on_unknown_or_incomplete(
        exception=ObjectivesError,
        known=[o.name for o in objectives],
        reported=evaluation.objectives.keys(),
    )


def call_functions_with_evaluations_and_collect_errors(
    functions: Iterable[Callable[[Evaluation], None]],
    evaluations: Iterable[Evaluation],
) -> None:
    """The given evaluations are passed to all given functions in order and the first
    Exception that occurrs for an evaluation is recorded alongside the evaluation and
    raised together with all erroneous evaluations as part of an `EvaluationsError`.

    NOTE: Even if reporting some evaluations fails, all that can be are successfully
    reported. Also, if an evaluation passes through some of the functions before causing
    issues, the effect the evaluation had on the previous functions can't be reverted.

    Raises:
        EvaluationsError: In case exceptions occurred when calling the functions with
            the evaluations.
    """
    evaluations_with_errors = []
    for evaluation in evaluations:
        for func in functions:
            try:
                func(evaluation)
            except EvaluationsError as e:
                evaluations_with_errors.extend(e.evaluations_with_errors)
                break
            except Exception as e:
                evaluations_with_errors.append((evaluation, e))
                break

    if evaluations_with_errors:
        raise EvaluationsError(evaluations_with_errors)


class Optimizer(abc.ABC):
    """Abstract base class for blackbox optimizer implementations."""

    def __init__(self, search_space: ParameterSpace, seed: int = None) -> None:
        """Initialize the optimizer with an optional seed for reproducibility.

        Args:
            search_space: The search space to optimize.
            seed: A seed for the optimizer, which is also used to re-seed the provided
                search space.
        """
        super().__init__()
        self.search_space = search_space
        self.seed = seed

        if self.seed is not None:
            self.search_space.seed(self.seed)

    @abc.abstractmethod
    def generate_evaluation_specification(self) -> EvaluationSpecification:
        """Get next configuration and settings to evaluate.

        Raises:
            OptimizationComplete: When the optimization run is finished, e.g. when the
                budget has been exhausted.
            OptimizerNotReady: When the optimizer is not ready to propose a new
                evaluation specification.
        """

    @abc.abstractmethod
    def report(self, evaluations: Union[Evaluation, Iterable[Evaluation]]) -> None:
        """Report one or more evaluated evaluation specifications.

        NOTE: Not all optimizers support reporting results for evaluation specifications
        that were not proposed by the optimizer.

        Args:
            evaluations: A single evaluated evaluation specifications, or an iterable
                of many.
        """


class SingleObjectiveOptimizer(Optimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objective: Objective,
        seed: int = None,
    ) -> None:
        """Initialize the optimizer with an optional seed for reproducibility.

        Args:
            search_space: The search space to optimize.
            objective: The objectives of the optimization.
            seed: A seed for the optimizer, which is also used to re-seed the provided
                search space.
        """
        super().__init__(search_space=search_space, seed=seed)
        self.objective = objective

    def report(self, evaluations: Union[Evaluation, Iterable[Evaluation]]) -> None:
        """Report one or multiple evaluations to the optimizer.

        All valid evaluations are processed. Faulty evaluations are not processed,
        instead an `EvaluationsError` is raised, which includes the problematic
        evaluations with their respective Exceptions in the `evaluations_with_errors`
        attribute.

        Args:
            evaluations: A single evaluated evaluation specifications, or an iterable
                of many.

        Raises:
            EvaluationsError: Raised when an evaluation could not be processed.
        """
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations

        call_functions_with_evaluations_and_collect_errors(
            [functools.partial(validate_objectives, objectives=[self.objective])],
            _evals,
        )


class MultiObjectiveOptimizer(Optimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objectives: List[Objective],
        seed: int = None,
    ) -> None:
        """Initialize the optimizer with an optional seed for reproducibility.

        Args:
            search_space: The search space to optimize.
            objectives: The objectives of the optimization.
            seed: A seed for the optimizer, which is also used to re-seed the provided
                search space.
        """
        super().__init__(search_space=search_space, seed=seed)

        _raise_on_duplicate_objective_names(objectives)
        self.objectives = objectives

    def report(self, evaluations: Union[Evaluation, Iterable[Evaluation]]) -> None:
        _evals = [evaluations] if isinstance(evaluations, Evaluation) else evaluations

        call_functions_with_evaluations_and_collect_errors(
            [functools.partial(validate_objectives, objectives=self.objectives)],
            _evals,
        )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import parameterspace as ps

try:
    import dask.distributed as dd
except ImportError:
    raise ImportError(
        "Unable to import Dask Distributed specific dependencies. "
        + "Make sure to install blackboxopt[dask]"
    )

from blackboxopt import Evaluation, EvaluationSpecification, Objective
from blackboxopt.optimization_loops.dask_distributed import (
    run_optimization_loop,
)
from blackboxopt.optimizers.random_search import RandomSearch


def evaluation_function(eval_spec: EvaluationSpecification) -> Evaluation:
    return eval_spec.create_evaluation(
        objectives={"loss": eval_spec.configuration["p1"] ** 2},
        user_info={"weather": "sunny"},
    )


if __name__ == "__main__":
    space = ps.ParameterSpace()
    space.add(ps.ContinuousParameter("p1", (-1.0, 1.0)))
    optimizer = RandomSearch(
        space,
        [Objective("loss", greater_is_better=False)],
        max_steps=1000,
    )

    evaluations = run_optimization_loop(
        optimizer, evaluation_function, dd.Client(), max_evaluations=100
    )

    n_successes = len([e for e in evaluations if not e.all_objectives_none])
    print(f"Successfully evaluated {n_successes}/{len(evaluations)}")



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0
#
# This source code is derived from HpBandSter 0.7.4
#  https://github.com/automl/HpBandSter
# Copyright (c) 2017-2018, ML4AAD, licensed under the BSD-3 license,
# cf. 3rd-party-licenses.txt file in the root directory of this source tree.
# Changes include:
#     - integration into the blackboxopt API
#     - move code into functions instead of class methods
#     - docstrings and type hints

import logging
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import parameterspace as ps
import scipy.stats as sps
import statsmodels.api as sm
from parameterspace import ParameterSpace

from blackboxopt import Evaluation, Objective
from blackboxopt.optimizers.staged.configuration_sampler import (
    StagedIterationConfigurationSampler,
)


def sample_around_values(
    datum: np.ndarray,
    bandwidths: np.ndarray,
    vartypes: Union[list, np.ndarray],
    min_bandwidth: float,
    bw_factor: float,
    rng: Optional[np.random.Generator] = None,
) -> Optional[np.ndarray]:
    """Sample numerical representation close to a given datum.

    This is specific to the KDE in statsmodels and their kernel for the different
    variable types.

    Args:
        datum: Numerical representation of a configuration that is used as the 'center'
            for sampling.
        bandwidths: Bandwidth of the corresponding kernels in each dimension.
        vartypes: Encoding of the types of the variables: 0 mean continuous, >0 means
            categorical with as many different values.
        min_bandwidth: Smallest allowed bandwidth. Ensures diversity even if all
            samples agree on a value in a dimension.
        bw_factor: To increase diversity, the bandwidth is actually multiplied by this
            factor before sampling.
        rng: A random number generator to make the sampling reproducible.

    Returns:
        Numerical representation of a configuration close to the provided datum.
    """
    rng = np.random.default_rng(rng)

    vector = []
    for m, bw, t in zip(datum, bandwidths, vartypes):
        bw = max(bw, min_bandwidth)
        if t == 0:
            bw = bw_factor * bw
            try:
                v = sps.truncnorm.rvs(
                    -m / bw, (1 - m) / bw, loc=m, scale=bw, random_state=rng
                )
            except Exception:
                return None
        elif t > 0:
            v = m if rng.random() < (1 - bw) else rng.integers(t)
        else:
            bw = min(0.9999, bw)  # bandwidth has to be less the one for this kernel!
            diffs = np.abs(np.arange(-t) - m)
            probs = 0.5 * (1 - bw) * (bw**diffs)
            idx = diffs == 0
            probs[idx] = (idx * (1 - bw))[idx]
            probs /= probs.sum()
            v = rng.choice(-t, p=probs)
        vector.append(v)
    return np.array(vector)


def convert_to_statsmodels_kde_representation(
    array: np.ndarray, vartypes: Union[list, np.ndarray]
) -> np.ndarray:
    """Convert numerical representation for categoricals and ordinals to integers.
    Args:
        array: Numerical representation of the configurations with categorical and ordinal values
            mapped into the unit hypercube.
        vartypes: Encoding of the types of the variables: 0 mean continuous, >0 means
            categorical with as many different values, and <0 means ordinal with as many values.

    Returns:
        Numerical representation consistent with the statsmodels package.
    """
    processed_vector = np.copy(array)

    for i in range(len(processed_vector)):
        if vartypes[i] == 0:
            continue
        num_values = abs(vartypes[i])
        processed_vector[i] = np.around((processed_vector[i] * num_values) - 0.5)

    return processed_vector


def convert_from_statsmodels_kde_representation(
    array: np.ndarray, vartypes: Union[list, np.ndarray]
) -> np.ndarray:
    """Convert numerical representation for categoricals and ordinals back into the unit hypercube.

    Args:
        array: Numerical representation of the configurations following the statsmodels convention for
            categorical and ordinal values being integers.
        vartypes: Encoding of the types of the variables: 0 mean continuous, >0 means
            categorical with as many different values, and <0 means ordinal with as many values.

    Returns:
        Numerical representation consistent with a numerical representation in the hypercube.
    """
    processed_vector = np.copy(array)

    for i in range(len(processed_vector)):
        if vartypes[i] != 0:
            num_values = abs(vartypes[i])
            processed_vector[i] = (processed_vector[i] + 0.5) / num_values

    return processed_vector


def impute_conditional_data(
    array: np.ndarray,
    vartypes: Union[list, np.ndarray],
    rng: Optional[np.random.Generator] = None,
) -> np.ndarray:
    """Impute NaNs in numerical representation with observed values or prior samples.

    This method is needed to use the `statsmodels` KDE, which doesn't handle missing
    values out of the box.

    Args:
        array: Numerical representation of the configurations which can include NaN
            values for inactive variables.
        vartypes: Encoding of the types of the variables: 0 mean continuous, >0 means
            categorical with as many different values, and <0 means ordinal with as many values.
        rng: A random number generator to make the imputation reproducible.
    Returns:
        Numerical representation where all NaNs have been replaced with observed values
        or prior samples.
    """
    rng = np.random.default_rng(rng)

    return_array = np.empty_like(array)

    for i in range(array.shape[0]):
        datum = np.copy(array[i])
        nan_indices = np.argwhere(np.isnan(datum)).flatten()

        while np.any(nan_indices):
            nan_idx = nan_indices[0]
            valid_indices = np.argwhere(np.isfinite(array[:, nan_idx])).flatten()

            if len(valid_indices) > 0:
                # pick one of them at random and overwrite all NaN values
                row_idx = rng.choice(valid_indices)
                datum[nan_indices] = array[row_idx, nan_indices]

            else:
                # no point in the data has this value activated, so fill it with a valid
                # but random value
                t = vartypes[nan_idx]
                if t == 0:
                    datum[nan_idx] = rng.random()
                elif t > 0:
                    datum[nan_idx] = rng.integers(t)
                elif t < 0:
                    datum[nan_idx] = rng.integers(-t)
            nan_indices = np.argwhere(np.isnan(datum)).flatten()
        return_array[i, :] = datum
    return return_array


class Sampler(StagedIterationConfigurationSampler):
    def __init__(
        self,
        search_space: ParameterSpace,
        objective: Objective,
        min_samples_in_model: int,
        top_n_percent: int,
        num_samples: int,
        random_fraction: float,
        bandwidth_factor: float,
        min_bandwidth: float,
        seed: int = None,
        logger=None,
    ):
        """Fits for each given fidelity a kernel density estimator on the best N percent
        of the evaluated configurations on this fidelity.

        Args:
            search_space: ConfigurationSpace/ ParameterSpace object.
            objective: The objective of the optimization.
            min_samples_in_model: Minimum number of datapoints needed to fit a model.
            top_n_percent: Determines the percentile of configurations that will be used
                as training data for the kernel density estimator of the good
                configuration, e.g if set to 10 the best 10% configurations will be
                considered for training.
            num_samples: Number of samples drawn to optimize EI via sampling.
            random_fraction: Fraction of random configurations returned
            bandwidth_factor: Widens the bandwidth for contiuous parameters for
                proposed points to optimize EI
            min_bandwidth: To keep diversity, even when all (good) samples have the
                same value for one of the parameters, a minimum bandwidth
                (reasonable default: 1e-3) is used instead of zero.
            seed: A seed to make the sampler reproducible.
            logger: [description]

        Raises:
            RuntimeError: [description]
        """
        self.logger = logging.getLogger("blackboxopt") if logger is None else logger

        self.objective = objective
        self.min_samples_in_model = min_samples_in_model
        self.top_n_percent = top_n_percent
        self.search_space = search_space
        self.bw_factor = bandwidth_factor
        self.min_bandwidth = min_bandwidth
        self.seed = seed
        self._rng = np.random.default_rng(self.seed)

        if self.min_samples_in_model < len(search_space) + 1:
            self.min_samples_in_model = len(search_space) + 1
            self.logger.warning(
                "Invalid min_samples_in_model value. "
                + f"Setting it to {self.min_samples_in_model}"
            )

        self.num_samples = num_samples
        self.random_fraction = random_fraction

        self.kde_vartypes = ""

        vartypes: List[Union[float, int]] = []
        for hp in search_space:  # type: ignore
            hp = hp["parameter"]
            if isinstance(hp, (ps.ContinuousParameter, ps.IntegerParameter)):
                self.kde_vartypes += "c"
                vartypes.append(0)

            elif isinstance(hp, ps.CategoricalParameter):
                self.kde_vartypes += "u"
                vartypes.append(hp.num_values)

            elif isinstance(hp, ps.OrdinalParameter):
                self.kde_vartypes += "o"
                vartypes.append(-hp.num_values)
            else:
                raise RuntimeError(f"This version on BOHB does not support {type(hp)}!")

        self.vartypes = np.array(vartypes, dtype=int)

        self.configs: Dict[float, List[np.ndarray]] = dict()
        self.losses: Dict[float, List[float]] = dict()
        self.kde_models: Dict[float, dict] = dict()

    def sample_configuration(self) -> Tuple[dict, dict]:
        """[summary]

        Returns:
            [description]
        """
        self.logger.debug("start sampling a new configuration.")

        # Sample from prior, if no model is available or with given probability
        if len(self.kde_models) == 0 or self._rng.random() < self.random_fraction:
            return self.search_space.sample(), {"model_based_pick": False}

        best = np.inf
        best_vector = None

        try:
            # sample from largest fidelity
            fidelity = max(self.kde_models.keys())

            good = self.kde_models[fidelity]["good"].pdf
            bad = self.kde_models[fidelity]["bad"].pdf

            minimize_me = lambda x: max(1e-32, bad(x)) / max(good(x), 1e-32)

            kde_good = self.kde_models[fidelity]["good"]
            kde_bad = self.kde_models[fidelity]["bad"]

            for i in range(self.num_samples):
                idx = self._rng.integers(0, len(kde_good.data))
                datum = kde_good.data[idx]
                vector = sample_around_values(
                    datum,
                    kde_good.bw,
                    self.vartypes,
                    self.min_bandwidth,
                    self.bw_factor,
                    rng=self._rng,
                )
                if vector is None:
                    continue

                # Statsmodels KDE estimators relies on seeding through numpy's global
                # state. We do this close to the evaluation of the PDF (`good`, `bad`)
                # to increase robustness for multi threading.
                # As we seed in a loop, we need to change it each iteration to not get
                # the same random numbers each time.
                # We also reset the np.random's global state, in case the user relies
                # on it in other parts of the code and to not hide other determinism
                # issues.
                # TODO: Check github issue if there was progress and the seeding can be
                # removed: https://github.com/statsmodels/statsmodels/issues/306
                cached_rng_state = None
                if self.seed:
                    cached_rng_state = np.random.get_state()
                    np.random.seed(self.seed + i)

                val = minimize_me(vector)

                if cached_rng_state:
                    np.random.set_state(cached_rng_state)

                if not np.isfinite(val):
                    self.logger.warning(
                        "sampled vector: %s has EI value %s" % (vector, val)
                    )
                    self.logger.warning(
                        "data in the KDEs:\n%s\n%s" % (kde_good.data, kde_bad.data)
                    )
                    self.logger.warning(
                        "bandwidth of the KDEs:\n%s\n%s" % (kde_good.bw, kde_bad.bw)
                    )

                    # right now, this happens because a KDE does not contain all values
                    # for a categorical parameter this cannot be fixed with the
                    # statsmodels KDE, so for now, we are just going to evaluate this
                    # one if the good_kde has a finite value, i.e. there is no config
                    # with that value in the bad kde, so it shouldn't be terrible.
                    if np.isfinite(good(vector)) and best_vector is not None:
                        best_vector = vector
                    continue

                if val < best:
                    best = val
                    best_vector = convert_from_statsmodels_kde_representation(
                        vector, self.vartypes
                    )

            if best_vector is None:
                self.logger.debug(
                    f"Sampling based optimization with {self.num_samples} samples did "
                    + "not find any finite/numerical acquisition function value "
                    + "-> using random configuration"
                )
                return self.search_space.sample(), {"model_based_pick": False}
            else:
                self.logger.debug(
                    "best_vector: {}, {}, {}, {}".format(
                        best_vector, best, good(best_vector), bad(best_vector)
                    )
                )
                return (
                    self.search_space.from_numerical(best_vector),
                    {"model_based_pick": True},
                )

        except Exception as e:
            self.logger.debug(
                "Sample base optimization failed. Falling back to a random sample."
            )
            return self.search_space.sample(), {"model_based_pick": False}

    def digest_evaluation(self, evaluation: Evaluation):
        """[summary]

        Args:
            evaluation: [description]
        """
        objective_value = evaluation.objectives[self.objective.name]
        if objective_value is None:
            loss = np.inf
        else:
            loss = (
                -objective_value
                if self.objective.greater_is_better
                else objective_value
            )
        config_vector = self.search_space.to_numerical(evaluation.configuration)
        config_vector = convert_to_statsmodels_kde_representation(
            config_vector, self.vartypes
        )

        fidelity = evaluation.settings["fidelity"]

        if fidelity not in self.configs.keys():
            self.configs[fidelity] = []
            self.losses[fidelity] = []

        self.configs[fidelity].append(config_vector)
        self.losses[fidelity].append(loss)

        if bool(self.kde_models.keys()) and max(self.kde_models.keys()) > fidelity:
            return

        if np.isfinite(self.losses[fidelity]).sum() <= self.min_samples_in_model - 1:
            n_runs_finite_loss = np.isfinite(self.losses[fidelity]).sum()
            self.logger.debug(
                f"Only {n_runs_finite_loss} run(s) with a finite loss for fidelity "
                + f"{fidelity} available, need more than {self.min_samples_in_model+1} "
                + "-> can't build model!"
            )
            return

        train_configs = np.array(self.configs[fidelity])
        train_losses = np.array(self.losses[fidelity])

        n_good = max(
            self.min_samples_in_model,
            (self.top_n_percent * train_configs.shape[0]) // 100,
        )

        n_bad = max(
            self.min_samples_in_model,
            ((100 - self.top_n_percent) * train_configs.shape[0]) // 100,
        )

        # Refit KDE for the current fidelity
        idx = np.argsort(train_losses)

        train_data_good = impute_conditional_data(
            train_configs[idx[:n_good]], self.vartypes, rng=self._rng
        )
        train_data_bad = impute_conditional_data(
            train_configs[idx[n_good : n_good + n_bad]], self.vartypes, rng=self._rng
        )

        if train_data_good.shape[0] <= train_data_good.shape[1]:
            return
        if train_data_bad.shape[0] <= train_data_bad.shape[1]:
            return

        # more expensive crossvalidation method
        # bw_estimation = 'cv_ls'
        # quick rule of thumb
        bw_estimation = "normal_reference"

        bad_kde = sm.nonparametric.KDEMultivariate(
            data=train_data_bad,
            var_type=self.kde_vartypes,
            bw=bw_estimation,
        )
        good_kde = sm.nonparametric.KDEMultivariate(
            data=train_data_good,
            var_type=self.kde_vartypes,
            bw=bw_estimation,
        )

        bad_kde.bw = np.clip(bad_kde.bw, self.min_bandwidth, None)
        good_kde.bw = np.clip(good_kde.bw, self.min_bandwidth, None)

        self.kde_models[fidelity] = {"good": good_kde, "bad": bad_kde}

        # update probs for the categorical parameters for later sampling
        self.logger.debug(
            f"done building a new model for fidelity {fidelity} based on "
            + f"{n_good}/{n_bad} split\nBest loss for this fidelity: "
            + f"{np.min(train_losses)}\n"
            + ("=" * 40)
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blackboxopt',
 'blackboxopt.examples',
 'blackboxopt.optimization_loops',
 'blackboxopt.optimizers',
 'blackboxopt.optimizers.staged',
 'blackboxopt.visualizations']

package_data = \
{'': ['*']}

install_requires = \
['parameterspace>=0.7.2,<0.8.0']

extras_require = \
{'all': ['numpy>=1.22.0,<2.0.0',
         'plotly>=5.10.0,<6.0.0',
         'scipy>=1.6.0,<2.0.0',
         'statsmodels>=0.13.2,<0.14.0',
         'dask>=2021.2.0,<2022.0.0',
         'distributed>=2021.2.0,<2022.0.0',
         'pandas>=1.2.4,<2.0.0',
         'botorch>=0.7.1,<0.8.0'],
 'bohb': ['numpy>=1.22.0,<2.0.0',
          'scipy>=1.6.0,<2.0.0',
          'statsmodels>=0.13.2,<0.14.0'],
 'botorch': ['botorch>=0.7.1,<0.8.0'],
 'dask': ['dask>=2021.2.0,<2022.0.0', 'distributed>=2021.2.0,<2022.0.0'],
 'hyperband': ['numpy>=1.22.0,<2.0.0'],
 'space-fill': ['numpy>=1.22.0,<2.0.0', 'scipy>=1.6.0,<2.0.0'],
 'testing': ['numpy>=1.22.0,<2.0.0'],
 'visualization': ['plotly>=5.10.0,<6.0.0',
                   'scipy>=1.6.0,<2.0.0',
                   'pandas>=1.2.4,<2.0.0']}

setup_kwargs = {
    'name': 'blackboxopt',
    'version': '4.7.1',
    'description': 'A common interface for blackbox optimization algorithms along with useful helpers like parallel optimization loops, analysis and visualization scripts.',
    'long_description': '# Blackbox Optimization\n\n[![License](https://img.shields.io/badge/License-Apache%202.0-blue.svg)](LICENSE)\n[![CI/CD](https://github.com/boschresearch/blackboxopt/workflows/ci-cd-pipeline/badge.svg)](https://github.com/boschresearch/blackboxopt/actions?query=workflow%3Aci-cd-pipeline+branch%3Amain)\n\nVarious blackbox optimization algorithms with a common interface along with useful\nhelpers like parallel optimization loops, analysis and visualization scripts.\n\nRandom search is provided as an example optimizer along with tests for the interface.\n\nNew optimizers can require `blackboxopt` as a dependency, which is just the light-weight\ninterface definition.\nIf you want all optimizer implementations that come with this package, install\n`blackboxopt[all]`\nAlternatively, you can get individual optimizers with e.g. `blackboxopt[bohb]`\n\nThis software is a research prototype.\nThe software is not ready for production use.\nIt has neither been developed nor tested for a specific use case.\nHowever, the license conditions of the applicable Open Source licenses allow you to\nadapt the software to your needs.\nBefore using it in a safety relevant setting, make sure that the software fulfills your\nrequirements and adjust it according to any applicable safety standards\n(e.g. ISO 26262).\n\n## Documentation\n\n**Visit [boschresearch.github.io/blackboxopt](https://boschresearch.github.io/blackboxopt/)**\n\n## Development\n\nInstall poetry\n\n```\npip install poetry\n```\n\nInstall the `blackboxopt` package from source by running the following from the root\ndirectory of _this_ repository\n\n```\npoetry install\n```\n\n(Optional) Install [pre-commit](https://pre-commit.com) hooks to check code standards\nbefore committing changes:\n\n```\npoetry run pre-commit install\n```\n\n## Test\n\nMake sure to install all extras before running tests\n\n```\npoetry install -E testing\npoetry run pytest tests/\n```\n\nFor HTML test coverage reports run\n\n```\npoetry run pytest tests/ --cov --cov-report html:htmlcov\n```\n\n### Custom Optimizers\n\nWhen you develop an optimizer based on the interface defined as part of\n`blackboxopt.base`, you can use `blackboxopt.testing` to directly test whether your\nimplementation follows the specification by adding a test like this to your test suite.\n\n```python\nfrom blackboxopt.testing import ALL_REFERENCE_TESTS\n\n@pytest.mark.parametrize("reference_test", ALL_REFERENCE_TESTS)\ndef test_all_reference_tests(reference_test):\n    reference_test(CustomOptimizer, custom_optimizer_init_kwargs)\n```\n\n## Building Documentation\n\nMake sure to install _all_ necessary dependencies:\n\n```\npoetry install --extras=all\n```\n\nThe documentation can be built from the repository root as follows:\n\n```\npoetry run mkdocs build --clean --no-directory-urls\n```\n\nFor serving it locally while working on the documentation run:\n\n```\npoetry run mkdocs serve\n```\n\n## Architectural Decision Records\n\n### Create evaluation result from specification\n\nIn the context of initializing an evaluation result from a specification, facing the\nconcern that having a constructor with a specification argument while the specification\nattributes end up as toplevel attributes and not summarized under a specification\nattribute we decided for unpacking the evaluation specification like a dictionary into\nthe result constructor to prevent the said cognitive dissonance, accepting that the\nunpacking operator can feel unintuitive and that users might tend to matching the\nattributes explictly to the init arguments.\n\n### Report multiple evaluations\n\nIn the context of many optimizers just sequentally reporting the individual evaluations\nwhen multiple evaluations are reported at once and thus not leveraging any batch\nreporting benefits, facing the concern that representing that common behaviour in the\noptimizer base class requires the definition of an abstract report single and an\nabstract report multi method for which the report single does not need to be implemented\nif the report multi is, we decided to refactor the arising redundancy into a function\n`call_functions_with_evaluations_and_collect_errors`, accepting that this increases the\ncognitive load when reading the code.\n\n## License\n\n`blackboxopt` is open-sourced under the Apache-2.0 license. See the [LICENSE](LICENSE)\nfile for details.\n\nFor a list of other open source components included in `blackboxopt`, see the file\n[3rd-party-licenses.txt](3rd-party-licenses.txt).\n',
    'author': 'Bosch Center for AI, Robert Bosch GmbH',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/boschresearch/blackboxopt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import logging
import math

from blackboxopt.base import Objective
from blackboxopt.optimizers.staged.configuration_sampler import (
    StagedIterationConfigurationSampler,
)
from blackboxopt.optimizers.staged.iteration import StagedIteration
from blackboxopt.optimizers.staged.utils import greedy_promotion


# The following function was derived from HpBandSter 0.7.4
#  https://github.com/automl/HpBandSter
# Copyright (c) 2017-2018, ML4AAD, licensed under the BSD-3 license,
# cf. 3rd-party-licenses.txt file in the root directory of this source tree.
# The following changes have been made:
# - code was part of a method, which was moved into this function
# - docstring and type hints for the arguments were added
# - integration into the local classes, e.g.
#   blackboxopt.optimizers.staged.iteration.StagetIteration
def create_hyperband_iteration(
    iteration_index: int,
    min_fidelity: float,
    max_fidelity: float,
    eta: float,
    config_sampler: StagedIterationConfigurationSampler,
    objective: Objective,
    logger: logging.Logger,
) -> StagedIteration:
    """Optimizer specific way to create a new
    `blackboxopt.optimizer.staged.iteration.StagedIteration` object
    """
    # 's_max + 1' in the paper
    max_num_stages = 1 + int(math.floor(math.log(max_fidelity / min_fidelity, eta)))
    # 's+1' in the paper
    num_stages = max_num_stages - (iteration_index % (max_num_stages))
    num_configs_first_stage = int(
        math.ceil((max_num_stages / num_stages) * eta ** (num_stages - 1))
    )
    num_configs_per_stage = [
        int(num_configs_first_stage // (eta**i)) for i in range(num_stages)
    ]
    fidelities_per_stage = [
        max_fidelity / eta**i for i in range(num_stages - 1, -1, -1)
    ]
    # Hyperband simple draws random configurations, and there is no additional
    # information that needs to be stored
    return StagedIteration(
        iteration_index,
        num_configs_per_stage,
        fidelities_per_stage,
        config_sampler,
        greedy_promotion,
        objective,
        logger=logger,
    )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0
#
# This source code is derived from HpBandSter 0.7.4
#  https://github.com/automl/HpBandSter
# Copyright (c) 2017-2018, ML4AAD, licensed under the BSD-3 license,
# cf. 3rd-party-licenses.txt file in the root directory of this source tree.
# Changes include:
#     - integration into the blackboxopt API
#     - docstrings and type hints

import logging

from parameterspace import ParameterSpace

from blackboxopt import Objective

try:
    from blackboxopt.optimizers.staged.bohb import Sampler as BOHBSampler
    from blackboxopt.optimizers.staged.hyperband import create_hyperband_iteration
    from blackboxopt.optimizers.staged.optimizer import StagedIterationOptimizer
except ImportError as e:
    raise ImportError(
        "Unable to import BOHB optimizer specific dependencies. "
        + "Make sure to install blackboxopt[bohb]"
    ) from e


class BOHB(StagedIterationOptimizer):
    def __init__(
        self,
        search_space: ParameterSpace,
        objective: Objective,
        min_fidelity: float,
        max_fidelity: float,
        num_iterations: int,
        eta: float = 3.0,
        top_n_percent: int = 15,
        min_samples_in_model: int = None,
        num_samples: int = 64,
        random_fraction: float = 1 / 3,
        bandwidth_factor: float = 3.0,
        min_bandwidth: float = 1e-3,
        seed: int = None,
        logger: logging.Logger = None,
    ):
        """BOHB Optimizer.

        BOHB performs robust and efficient hyperparameter optimization
        at scale by combining the speed of Hyperband searches with the
        guidance and guarantees of convergence of Bayesian
        Optimization. Instead of sampling new configurations at random,
        BOHB uses kernel density estimators to select promising candidates.

        For reference:
        ```
        @InProceedings{falkner-icml-18,
            title =     {{BOHB}: Robust and Efficient Hyperparameter Optimization at
                Scale},
            author =    {Falkner, Stefan and Klein, Aaron and Hutter, Frank},
            booktitle = {Proceedings of the 35th International Conference on Machine
                Learning},
            pages =     {1436--1445},
            year =      {2018},
        }
        ```

        Args:
            search_space: [description]
            objective: [description]
            min_fidelity: The smallest fidelity value that is still meaningful.
                Must be strictly greater than zero!
            max_fidelity: The largest fidelity value used during the optimization.
                Must not be smaller than `min_fidelity`.
            num_iterations: The number of iterations that the optimizer will run.
            eta: Scaling parameter to control the aggressiveness of Hyperband's racing.
            top_n_percent: Determines the percentile of configurations that will be
                used as training data for the kernel density estimator of the good
                configuration, e.g if set to 10 the best 10% configurations will be
                considered for training.
            min_samples_in_model: Minimum number of datapoints needed to fit a model.
            num_samples: Number of samples drawn to optimize EI via sampling.
            random_fraction: Fraction of random configurations returned.
            bandwidth_factor: Widens the bandwidth for contiuous parameters for
                proposed points to optimize EI
            min_bandwidth: to keep diversity, even when all (good) samples have the
                same value for one of the parameters, a minimum bandwidth
                (reasonable default: 1e-3) is used instead of zero.
            seed: [description]
            logger: [description]
        """
        if min_samples_in_model is None:
            min_samples_in_model = 3 * len(search_space)

        self.min_fidelity = min_fidelity
        self.max_fidelity = max_fidelity
        self.eta = eta

        self.config_sampler = BOHBSampler(
            search_space=search_space,
            objective=objective,
            min_samples_in_model=min_samples_in_model,
            top_n_percent=top_n_percent,
            num_samples=num_samples,
            random_fraction=random_fraction,
            bandwidth_factor=bandwidth_factor,
            min_bandwidth=min_bandwidth,
            seed=seed,
        )

        super().__init__(
            search_space=search_space,
            objective=objective,
            num_iterations=num_iterations,
            seed=seed,
            logger=logger,
        )

    def _create_new_iteration(self, iteration_index):
        """Optimizer specific way to create a new
        `blackboxopt.optimizer.utils.staged_iteration.StagedIteration` object
        """
        return create_hyperband_iteration(
            iteration_index,
            self.min_fidelity,
            self.max_fidelity,
            self.eta,
            self.config_sampler,
            self.objective,
            self.logger,
        )



# Copyright (c) 2020 - for information on the respective copyright owner
# see the NOTICE file and/or the repository https://github.com/boschresearch/blackboxopt
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import time
from typing import Any, Callable, List, Optional, Union

from blackboxopt import (
    Evaluation,
    EvaluationSpecification,
    OptimizationComplete,
    OptimizerNotReady,
)
from blackboxopt.base import MultiObjectiveOptimizer, SingleObjectiveOptimizer
from blackboxopt.optimization_loops.utils import (
    evaluation_function_wrapper,
    init_max_evaluations_with_limit_logging,
)


def run_optimization_loop(
    optimizer: Union[SingleObjectiveOptimizer, MultiObjectiveOptimizer],
    evaluation_function: Callable[[EvaluationSpecification], Evaluation],
    timeout_s: float = float("inf"),
    max_evaluations: int = None,
    catch_exceptions_from_evaluation_function: bool = False,
    post_evaluation_callback: Optional[Callable[[Evaluation], Any]] = None,
    logger: logging.Logger = None,
) -> List[Evaluation]:
    """Convenience wrapper for an optimization loop that sequentially fetches evaluation
    specifications until a given timeout or maximum number of evaluations is reached.

    This already handles signals from the optimizer in case there is no evaluation
    specification available yet.

    Args:
        optimizer: The blackboxopt optimizer to run.
        evaluation_function: The function that is called with configuration, settings
            and optimizer info dictionaries as arguments like provided by an evaluation
            specification.
            This is the function that encapsulates the actual execution of
            a parametrized experiment (e.g. ML model training) and should return a
            `blackboxopt.Evaluation` as a result.
        timeout_s: If given, the optimization loop will terminate after the first
            optimization step that exceeded the timeout (in seconds). Defaults to inf.
        max_evaluations: If given, the optimization loop will terminate after the given
            number of steps.
        catch_exceptions_from_evaluation_function: Whether to exit on an unhandled
            exception raised by the evaluation function or instead store their stack
            trace in the evaluation's `stacktrace` attribute. Set to True if there are
            spurious errors due to e.g. numerical instability that should not halt the
            optimization loop.
        post_evaluation_callback: Reference to a callable that is invoked after each
            evaluation and takes a `blackboxopt.Evaluation` as its argument.
        logger: The logger to use for logging progress.

    Returns:
        List of evluation specification and result for all evaluations.
    """
    logger = logging.getLogger("blackboxopt") if logger is None else logger

    objectives = (
        optimizer.objectives
        if isinstance(optimizer, MultiObjectiveOptimizer)
        else [optimizer.objective]
    )
    evaluations: List[Evaluation] = []

    _max_evaluations = init_max_evaluations_with_limit_logging(
        max_evaluations=max_evaluations, timeout_s=timeout_s, logger=logger
    )

    start = time.time()
    num_evaluations = 0
    while time.time() - start < timeout_s and num_evaluations < _max_evaluations:
        num_evaluations += 1

        try:
            evaluation_specification = optimizer.generate_evaluation_specification()
            logger.info(
                "The optimizer proposed a specification for evaluation:\n"
                + f"{json.dumps(evaluation_specification.to_dict(), indent=2)}"
            )

            evaluation = evaluation_function_wrapper(
                evaluation_function=evaluation_function,
                evaluation_specification=evaluation_specification,
                logger=logger,
                objectives=objectives,
                catch_exceptions_from_evaluation_function=catch_exceptions_from_evaluation_function,
            )
            logger.info(
                "Reporting the result from the evaluation function to the optimizer:\n"
                + f"{json.dumps(evaluation.to_dict(), indent=2)}"
            )
            optimizer.report(evaluation)
            evaluations.append(evaluation)

            if post_evaluation_callback is not None:
                post_evaluation_callback(evaluation)

        except OptimizerNotReady:
            logger.info("Optimizer is not ready yet, retrying in two seconds")
            time.sleep(2)
            continue

        except OptimizationComplete:
            logger.info("Optimization is complete")
            return evaluations

    logger.info("Aborting optimization due to specified maximum evaluations or timeout")
    return evaluations
